import wrapt

from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.mysql import MysqlIntegration

mysql_integration = MysqlIntegration()


class MysqlCursorWrapper(wrapt.ObjectProxy):

    def __init__(self, cursor, connection_wrapper):
        super(MysqlCursorWrapper, self).__init__(cursor)
        self._self_connection = connection_wrapper

    def execute(self, *args, **kwargs):
        return mysql_integration.run_and_trace(
            self.__wrapped__.execute,
            self._self_connection,
            args,
            kwargs,
        )

    def callproc(self, *args, **kwargs):
        return mysql_integration.run_and_trace(
            self.__wrapped__.callproc,
            self._self_connection,
            args,
            kwargs,
        )

    def __enter__(self):
        # raise appropriate error if api not supported (should reach the user)
        self.__wrapped__.__enter__
        return self


class MysqlConnectionWrapper(wrapt.ObjectProxy):
    def cursor(self, *args, **kwargs):
        cursor = self.__wrapped__.cursor(*args, **kwargs)
        return MysqlCursorWrapper(cursor, self)


def _wrapper(wrapped, instance, args, kwargs):
    connection = wrapped(*args, **kwargs)
    return MysqlConnectionWrapper(connection)


def patch():
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_RDB_DISABLE):
        wrapt.wrap_function_wrapper(
            'mysql.connector',
            'connect',
            _wrapper)



from __future__ import absolute_import
import os
import traceback
from importlib import import_module

MODULES = {}
IGNORE_MODULES = ('__init__',)
PYTHON_EXTENSIONS = ('.py', '.pyc')

for module_name in os.listdir(os.path.dirname(__file__)):
    filename, ext = os.path.splitext(module_name)
    if filename in IGNORE_MODULES or ext not in PYTHON_EXTENSIONS:
        continue
    try:
        imported = import_module(".{}".format(filename), __name__)
        MODULES[filename] = imported
    except ImportError:
        traceback.print_exc()



from __future__ import division

import logging

from thundra import constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.context.execution_context_manager import ExecutionContextManager

logger = logging.getLogger(__name__)


class Resource:

    def __init__(self, span):
        self.type = span.class_name
        self.name = span.operation_name
        self.operation = str(span.get_tag(constants.SpanTags['OPERATION_TYPE']))
        self.count = 1
        self.error_count = 1 if span.erroneous() else 0
        self.security_violated_count = 1 if span.get_tag(constants.SecurityTags['VIOLATED']) else 0
        self.security_blocked_count = 1 if span.get_tag(constants.SecurityTags['BLOCKED']) else 0
        self.error_types = {span.get_tag('error.kind')} if span.erroneous() else set()
        self.duration = span.get_duration()
        self.resource_max_duration = self.duration
        self.resource_trace_links = set()
        if hasattr(span, 'resource_trace_links') and hasattr(span.resource_trace_links, '__iter__'):
            self.resource_trace_links = set(span.resource_trace_links)

    def accept(self, span):
        return (
                self.type.upper() == span.class_name.upper() and
                self.name == span.operation_name and
                self.operation == str(span.get_tag(constants.SpanTags['OPERATION_TYPE']))
        )

    def merge(self, span):
        if not self.accept(span):
            logger.error(("can not merge resource with id %s:"
                          " ids not match with target resource"), _resource_id(span))
            return
        self.count += 1
        self.duration += span.get_duration()
        errorneous = span.erroneous()

        if span.get_tag(constants.SecurityTags['VIOLATED']):
            self.security_violated_count += 1

        if span.get_tag(constants.SecurityTags['BLOCKED']):
            self.security_blocked_count += 1

        if errorneous:
            self.error_count += 1
            self.error_types.add(span.get_tag('error.kind'))

        if span.get_duration() > self.resource_max_duration:
            self.resource_max_duration = span.get_duration()

        if self.resource_trace_links and hasattr(span, 'resource_trace_links') and hasattr(span.resource_trace_links,
                                                                                           '__iter__'):
            self.resource_trace_links.update(span.resource_trace_links)

    def to_dict(self):
        resource = {
            'resourceType': self.type,
            'resourceName': self.name,
            'resourceOperation': self.operation,
            'resourceCount': self.count,
            'resourceErrorCount': self.error_count,
            'resourceViolatedCount': self.security_violated_count,
            'resourceBlockedCount': self.security_blocked_count,
            'resourceDuration': self.duration,
            'resourceErrors': list(self.error_types),
            'resourceMaxDuration': self.resource_max_duration,
            'resourceAvgDuration': self.duration / self.count
        }
        if self.resource_trace_links:
            resource['resourceTraceLinks'] = list(self.resource_trace_links)
        return resource


def _resource_id(span, resource_name=None):
    if resource_name:
        return ('{}${}${}'.format(
            span.class_name.upper(),
            resource_name,
            str(span.get_tag(constants.SpanTags['OPERATION_TYPE']))
        ))
    return ('{}${}${}'.format(
        span.class_name.upper(),
        span.operation_name,
        str(span.get_tag(constants.SpanTags['OPERATION_TYPE']))
    ))


def get_resources():
    try:
        resources = {}
        execution_context = ExecutionContextManager.get()
        spans = execution_context.recorder.get_spans()
        root_span_id = execution_context.root_span.context.span_id
        for span in spans:
            if (not span.get_tag(constants.SpanTags['TOPOLOGY_VERTEX'])
                    or span.span_id == root_span_id):
                continue

            resource_names = span.get_tag(constants.SpanTags['RESOURCE_NAMES'])
            if resource_names:
                for resource_name in resource_names:
                    rid = _resource_id(span, resource_name)
                    if rid:
                        if not rid in resources:
                            resources[rid] = Resource(span)
                            resources[rid].name = resource_name
                        else:
                            resources[rid].merge(span)
            else:
                rid = _resource_id(span)
                if rid:
                    if not rid in resources:
                        resources[rid] = Resource(span)
                    else:
                        resources[rid].merge(span)

        return {
            'resources': [r.to_dict() for r in resources.values()]
        }
    except Exception as e:
        logger.error("error while creating the resources data for invocation: %s", e)
        return {}


def get_incoming_trace_links():
    if ConfigProvider.get(config_names.THUNDRA_DISABLE, False):
        return {}

    execution_context = ExecutionContextManager.get()
    incoming_trace_links = list(set(execution_context.incoming_trace_links))[:constants.MAX_INCOMING_TRACE_LINKS]
    return {"incomingTraceLinks": incoming_trace_links}


def get_outgoing_trace_links():
    if ConfigProvider.get(config_names.THUNDRA_DISABLE, False):
        return {}

    execution_context = ExecutionContextManager.get()
    spans = execution_context.recorder.get_spans()

    outgoing_trace_links = []
    outgoing_trace_links.extend(execution_context.outgoing_trace_links)
    for span in spans:
        links = get_outgoing_trace_link(span)
        if links:
            outgoing_trace_links += links

    outgoing_trace_links = list(set(outgoing_trace_links))[:constants.MAX_OUTGOING_TRACE_LINKS]
    return {"outgoingTraceLinks": outgoing_trace_links}


def add_outgoing_trace_link(trace_link):
    execution_context = ExecutionContextManager.get()
    execution_context.outgoing_trace_links.append(trace_link)


def add_outgoing_trace_links(trace_links):
    execution_context = ExecutionContextManager.get()
    execution_context.outgoing_trace_links.extend(trace_links)


def get_outgoing_trace_link(span):
    return span.get_tag(constants.SpanTags["TRACE_LINKS"])


def add_incoming_trace_link(trace_link):
    execution_context = ExecutionContextManager.get()
    execution_context.incoming_trace_links.append(trace_link)


def add_incoming_trace_links(trace_links):
    execution_context = ExecutionContextManager.get()
    execution_context.incoming_trace_links.extend(trace_links)



from __future__ import absolute_import

from importlib import import_module as _import_module

import thundra.integrations.modules as integrations
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.opentracing.tracer import ThundraTracer
from thundra.plugins.invocation.invocation_support import (
    set_tag,
    set_tags,
    remove_tag,
    get_tag,
    get_tags,
    set_error,
    clear_error
)
from thundra.plugins.invocation.invocation_trace_support import (
    add_incoming_trace_link,
    add_incoming_trace_links,
    add_outgoing_trace_link,
    add_outgoing_trace_links,
    get_incoming_trace_links,
    get_outgoing_trace_links,
)
from thundra.plugins.log import log_support as log
from thundra.plugins.metric import metric_support as metric
from thundra.plugins.trace import trace_support as trace
from thundra.plugins.trace.trace_aware_wrapper import TraceAwareWrapper
from thundra.plugins.trace.traceable import Traceable
from thundra.wrappers.wrapper_factory import WrapperFactory as _WrapperFactory

import logging
logger = logging.getLogger(__name__)

def _import_exists(module_name):
    try:
        _import_module(module_name)
        return True
    except ImportError:
        return False


def _patch_modules():
    for module_name, module in integrations.MODULES.items():
        if _import_exists(module_name):
            if hasattr(module, "patch"):
                try:
                    module.patch()
                except Exception as e:
                    logger.error("Couldn't patch module: %s", e)


def configure(options=None):
    ConfigProvider.__init__(options)


def lambda_wrapper(func):
    from thundra.wrappers.aws_lambda.lambda_wrapper import LambdaWrapper
    return _WrapperFactory.get_or_create(LambdaWrapper)(func)


def django_wrapper(func):
    from thundra.wrappers.django.django_wrapper import DjangoWrapper
    return _WrapperFactory.get_or_create(DjangoWrapper)(func)


def flask_wrapper(func):
    from thundra.wrappers.flask.flask_wrapper import FlaskWrapper
    return _WrapperFactory.get_or_create(FlaskWrapper)(func)


def fastapi_wrapper(func):
    from thundra.wrappers.fastapi.fastapi_wrapper import FastapiWrapper
    return _WrapperFactory.get_or_create(FastapiWrapper)(func)


def tornado_wrapper(func):
    from thundra.wrappers.tornado.tornado_wrapper import TornadoWrapper
    return _WrapperFactory.get_or_create(TornadoWrapper)(func)


def _set_thundra_for_test_env(already_configured):
    if not already_configured:
        configure()
    if not ConfigProvider.get(config_names.THUNDRA_DISABLE):
        _patch_modules()


if not ConfigProvider.get(config_names.THUNDRA_DISABLE):
    test_active = ConfigProvider.get(config_names.THUNDRA_TEST_ACTIVE)
    if not test_active:
        _patch_modules()


__all__ = [
    'configure',
    'TraceAwareWrapper',
    'log',
    'metric',
    'trace',
    'lambda_wrapper',
    'django_wrapper',
    'flask_wrapper',
    'fastapi_wrapper',
    'tornado_wrapper',
    'add_incoming_trace_link',
    'add_incoming_trace_links',
    'add_outgoing_trace_link',
    'add_outgoing_trace_links',
    'get_incoming_trace_links',
    'get_outgoing_trace_links',
    'set_tag',
    'set_tags',
    'remove_tag',
    'get_tag',
    'get_tags',
    'set_error',
    'clear_error',
    'Traceable'
]



from __future__ import absolute_import
import logging
from threading import Lock
from importlib import import_module
from thundra import constants
from thundra.listeners.thundra_span_listener import ThundraSpanListener

logger = logging.getLogger(__name__)
default_error_message = "Error injected by Thundra!"
default_error_type = Exception

try:
    import builtins
except:
    import __builtin__ as builtins

class ErrorInjectorSpanListener(ThundraSpanListener):
    def __init__(self, error_message=default_error_message,
                 error_type=default_error_type, inject_on_finish=False,
                 inject_count_freq=1, add_info_tags=True):

        self._counter = 0
        self._lock = Lock()

        self.error_message = error_message
        self.error_type = error_type
        self.inject_on_finish = inject_on_finish
        self.inject_count_freq = max(inject_count_freq, 1)
        self.add_info_tags = add_info_tags

    def on_span_started(self, span):
        if (not self.inject_on_finish
                and self._able_to_raise()):
            self._raise_error(span)

    def on_span_finished(self, span):
        if (self.inject_on_finish
                and self._able_to_raise()):
            self._raise_error(span)

    def _raise_error(self, span):
        err = self.error_type(self.error_message)
        if self.add_info_tags:
            self._add_info_tags(span)
        raise err

    def _able_to_raise(self):
        if self._increment_and_get_counter() % self.inject_count_freq == 0:
            return True
        return False

    def _increment_and_get_counter(self):
        with self._lock:
            self._counter += 1

        return self._counter

    def _add_info_tags(self, span):
        try:
            info_dict = {
                'type': 'error_injector_span_listener',
                'error_type': str(self.error_type),
                'error_message': self.error_message,
                'inject_on_finish': self.inject_on_finish,
                'inject_count_freq': self.inject_count_freq,
            }
            span.set_tag(constants.THUNDRA_LAMBDA_SPAN_LISTENER_INFO_TAG, info_dict)
        except Exception as e:
            logger.error("error while adding ErrorInjectorSpanListener info tags: %s", e)

    @staticmethod
    def from_config(config):
        kwargs = {}
        error_message = config.get('errorMessage')
        error_type = config.get('errorType')
        inject_on_finish = config.get('injectOnFinish')
        inject_count_freq = config.get('injectCountFreq')
        add_info_tags = config.get('addInfoTags')

        if error_message is not None:
            try:
                kwargs['error_message'] = error_message.strip('"')
            except AttributeError:
                ErrorInjectorSpanListener._log_value_parse_err(error_message, 'error_message')
        if error_type is not None:
            if hasattr(builtins, error_type):
                err_class = getattr(builtins, error_type)
                if issubclass(err_class, BaseException):
                    kwargs['error_type'] = getattr(builtins, error_type)
            else:
                try:
                    (err_module_name, err_class_name) = error_type.rsplit('.', 1)
                    err_module = import_module(err_module_name)
                    if hasattr(err_module, err_class_name):
                        err_class = getattr(err_module, err_class_name)
                        if issubclass(err_class, BaseException):
                            kwargs['error_type'] = err_class
                except (ImportError, ValueError):
                    logger.error("couldn't import %s", error_type)
        if inject_on_finish is not None:
            try:
                kwargs['inject_on_finish'] = inject_on_finish
            except ValueError:
                ErrorInjectorSpanListener._log_value_parse_err(inject_on_finish, 'inject_on_finish')
        if inject_count_freq is not None:
            try:
                kwargs['inject_count_freq'] = int(inject_count_freq)
            except ValueError:
                ErrorInjectorSpanListener._log_value_parse_err(inject_count_freq, 'inject_count_freq')
        if add_info_tags is not None:
            try:
                kwargs['add_info_tags'] = add_info_tags
            except ValueError:
                ErrorInjectorSpanListener._log_value_parse_err(add_info_tags, 'add_info_tags')

        return ErrorInjectorSpanListener(**kwargs)

    @staticmethod
    def _log_value_parse_err(param, param_name):
        logger.error(("couldn't parse %s parameter (%s) of "
                      "ErrorInjectorSpanListener, using the default value"), param_name, param)

    @staticmethod
    def should_raise_exceptions():
        return True



from abc import ABCMeta, abstractmethod

ABC = ABCMeta('ABC', (object,), {})


class BaseSampler(ABC):

    @abstractmethod
    def is_sampled(self, args):
        raise Exception("should be implemented")



import wrapt

from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.elasticsearch import ElasticsearchIntegration

es_integration = ElasticsearchIntegration()


def _wrapper(wrapped, instance, args, kwargs):
    return es_integration.run_and_trace(
        wrapped,
        instance,
        args,
        kwargs
    )


def patch():
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_ES_DISABLE):
        wrapt.wrap_function_wrapper(
            'elasticsearch',
            'transport.Transport.perform_request',
            _wrapper
        )



from thundra.application.application_info import ApplicationInfo
from thundra.application.application_info_provider import ApplicationInfoProvider
from foresight.test_runner_tags import TestRunnerTags
from foresight.utils.handler_utils import HandlerUtils
import foresight.pytest_integration.constants as pytest_constants
from foresight.constants.constants import ForesightConstants
import os, pytest, logging


logger = logging.getLogger(__name__)

class SpanManager:
    
    @staticmethod
    def handle_fixture_and_inject_span(handler, app_info=None, span_tags=None, request=None):
        """Handle beforeeach,aftereach, beforeall and afterall functions. Created span has been moved into request obj.

        Args:
            handler (Function): Beforeeach,aftereach, beforeall and afterall functions in utils.handler_utils
            app_info (ApplicationInfo, optional): Application Info. Defaults to None.
            span_tags (Dict, optional): [description]. Defaults to None.
            request (pytest.FixtureRequest, optional): Pytest FixtureRequest object that passed to wrapped fixture functions in pytest_integration.utils. Defaults to None.
        """
        try:
            scope = handler(app_info, span_tags)
            setattr(request, pytest_constants.THUNDRA_SCOPE, scope)
        except Exception as err:
            logger.error("Couldn't handle fixture and inject span for pytest: {}".format(err))
            pass


    @staticmethod
    def extract_scope(request):
        try:
            return getattr(request, pytest_constants.THUNDRA_SCOPE, None)
        except Exception as err:
            logger.error("Couldn't extract span from request for pytest: {}".format(err))
            pass
        return None


class PytestHelper:

    TEST_APP_ID_PREFIX = "python:test:pytest:"
    TEST_APP_VERSION = pytest.__version__
    TEST_CLASS_NAME = "Pytest"
    TEST_SUITE_PATH = 0
    TEST_OPERATION_NAME_INDEX = 2
    TEST_SUITE = "module"
    TEST_CASE = "function"
    PYTEST_STARTED = False
    PYTEST_COLLECT_ONLY = False


    @staticmethod
    def get_test_application_name(request):
        try:
            return request.nodeid.replace("::", os.sep)
        except Exception as err:
            logger.error("Couldn't get test application name for pytest: {}".format(err))
            pass
        return None

    @classmethod
    def get_test_application_id(cls, request):
        return cls.TEST_APP_ID_PREFIX + cls.get_test_application_name(request)


    @classmethod
    def get_test_application_instance_id(cls, request):
        return ForesightConstants.TEST_APP_INSTANCE_ID_PREFIX + cls.get_test_application_name(request)


    @classmethod
    def get_test_application_info(cls, request):
        try:
            domain_name = None
            application_id = None
            if request.scope == cls.TEST_SUITE:
                domain_name = ForesightConstants.TEST_SUITE_DOMAIN_NAME
                application_id = cls.get_test_application_id(request)
            elif request.scope == cls.TEST_CASE:
                domain_name = ForesightConstants.TEST_DOMAIN_NAME
            return ApplicationInfo(
                application_id,
                cls.get_test_application_instance_id(request),
                domain_name,
                cls.TEST_CLASS_NAME,
                cls.get_test_application_name(request),
                cls.TEST_APP_VERSION,
                ForesightConstants.TEST_APP_STAGE,
                ApplicationInfoProvider.APPLICATION_RUNTIME,
                ApplicationInfoProvider.APPLICATION_RUNTIME_VERSION,
                None,
            )
        except Exception as err:
            logger.error("Couldn't get application info for pytest: {}".format(err))
            pass

    @staticmethod
    def get_test_fixture_application_name(request):
        try:
            return request.fixturename
        except Exception as err:
            logger.error("Couldn't get fixture application name for pytest: {}".format(err))
            pass

    @classmethod
    def get_test_fixture_application_instance_id(cls, request):
        return ForesightConstants.TEST_APP_INSTANCE_ID_PREFIX + cls.get_test_fixture_application_name(request)


    @classmethod
    def get_test_fixture_application_info(cls, request):
        return ApplicationInfo(
            None,
            cls.get_test_fixture_application_instance_id(request),
            ForesightConstants.TEST_FIXTURE_DOMAIN_NAME,
            cls.TEST_CLASS_NAME,
            cls.get_test_fixture_application_name(request),
            cls.TEST_APP_VERSION,
            ForesightConstants.TEST_APP_STAGE,
            ApplicationInfoProvider.APPLICATION_RUNTIME,
            ApplicationInfoProvider.APPLICATION_RUNTIME_VERSION,
            None,
        )


    @classmethod
    def get_test_method_name(cls, request):
        try:
            nodeid = cls.get_test_application_name(request)
            if len(nodeid) > ForesightConstants.MAX_TEST_METHOD_NAME:
                nodeid = "..." + nodeid[(len(nodeid)-ForesightConstants.MAX_TEST_METHOD_NAME) + 3:]
            return nodeid
        except Exception as err:
            logger.error("Couldn't get method name for pytest: {}".format(err))
            pass
        return None 


    @classmethod
    def get_test_name(cls, request):
        return cls.get_test_method_name(request)


    @classmethod
    def check_pytest_started(cls):
        return cls.PYTEST_STARTED


    @classmethod
    def set_pytest_started(cls):
        cls.PYTEST_STARTED = True

    @classmethod
    def check_pytest_collect_only(cls):
        return cls.PYTEST_COLLECT_ONLY


    @classmethod
    def set_pytest_collect_only(cls):
        cls.PYTEST_COLLECT_ONLY = True

    @staticmethod
    def clear_test_case_state_for_thundra(item):
        delattr(item, pytest_constants.THUNDRA_TEST_RESULTED)
        delattr(item, pytest_constants.THUNDRA_TEST_STARTED)
        delattr(item, pytest_constants.THUNDRA_TEST_ALREADY_FINISHED)
        setattr(item, pytest_constants.THUNDRA_TEST_FINISH_IN_HELPER, True)
    

    @staticmethod
    def session_setup(executor, api_key=None):
        try:
            HandlerUtils.test_setup(executor, api_key)
        except Exception as err:
            logger.error("Couldn't setup the session for pytest: {}".format(err))
            pass


    @classmethod
    def session_teardown(cls):
        try:
            HandlerUtils.test_teardown()
        except Exception as err:
            logger.error("Error session teardown for pytest: {}".format(err))
            pass


    @classmethod
    def start_test_suite_span(cls, item):
        try:
            test_suite_id = item.nodeid
            app_info = cls.get_test_application_info(item)
            HandlerUtils.start_test_suite_span(test_suite_id, app_info)
        except Exception as err:
            logger.error("Couldn't start test suite span for pytest: {}".format(err))
            pass


    @classmethod
    def start_before_all_span(cls, request):
        try:
            app_info = cls.get_test_fixture_application_info(request).to_json()
            span_tags = {TestRunnerTags.TEST_SUITE: request.node.nodeid, TestRunnerTags.TEST_FIXTURE: request.fixturename}
            SpanManager.handle_fixture_and_inject_span(HandlerUtils.start_before_all_span, app_info, span_tags,
                request)
        except Exception as err:
            logger.error("Couldn't start before all span for pytest: {}".format(err))
            pass


    @staticmethod
    def finish_before_all_span(request):
        try:
            scope = SpanManager.extract_scope(request)
            HandlerUtils.finish_before_all_span(scope)
        except Exception as err:
            logger.error("Couldn't finish before all span for pytest: {}".format(err))
            pass


    @classmethod
    def start_after_all_span(cls, request):
        try:
            app_info = cls.get_test_fixture_application_info(request).to_json()
            span_tags = {TestRunnerTags.TEST_SUITE: request.node.nodeid, TestRunnerTags.TEST_FIXTURE: request.fixturename}
            SpanManager.handle_fixture_and_inject_span(HandlerUtils.start_after_all_span, app_info, span_tags,
                request)
        except Exception as err:
            logger.error("Couldn't start after all span for pytest: {}".format(err), err)
            pass


    @staticmethod
    def finish_after_all_span(request):
        try:
            scope = SpanManager.extract_scope(request)
            HandlerUtils.finish_after_all_span(scope)
        except Exception as err:
            logger.error("Couldn't finish after all span for pytest: {}".format(err))
            pass


    @staticmethod
    def finish_test_suite_span():
        try:
            HandlerUtils.finish_test_suite_span()
        except Exception as err:
            logger.error("Couldn't finish test suite span for pytest: {}".format(err))
            pass


    @classmethod
    def start_test_span(cls, item):
        try:
            if not hasattr(item, pytest_constants.THUNDRA_TEST_STARTED):
                setattr(item, pytest_constants.THUNDRA_TEST_STARTED, True)
                name = item.location[cls.TEST_OPERATION_NAME_INDEX]
                test_suite_name = item.location[cls.TEST_SUITE_PATH]
                test_case_id = item.nodeid
                app_info = cls.get_test_application_info(item)
                HandlerUtils.start_test_span(name, test_suite_name, test_case_id, app_info)
        except Exception as err:
            logger.error("Couldn't start test span for pytest: {}".format(err))
            pass


    @classmethod
    def start_before_each_span(cls, request):
        try:
            app_info = cls.get_test_fixture_application_info(request).to_json()
            span_tags = { TestRunnerTags.TEST_SUITE: request.node.location[cls.TEST_SUITE_PATH], TestRunnerTags.TEST_FIXTURE: request.fixturename}
            SpanManager.handle_fixture_and_inject_span(HandlerUtils.start_before_each_span, app_info, span_tags,
                request)
        except Exception as err:
            logger.error("Couldn't start before each span for pytest: {}".format(err))
            pass


    @staticmethod
    def finish_before_each_span(request):
        try:
            scope = SpanManager.extract_scope(request)
            HandlerUtils.finish_before_each_span(scope)
        except Exception as err:
            logger.error("Couldn't finish before each span for pytest: {}".format(err))
            pass


    @classmethod
    def start_after_each_span(cls, request):
        try:
            app_info = cls.get_test_fixture_application_info(request).to_json()
            span_tags = { TestRunnerTags.TEST_SUITE: request.node.location[cls.TEST_SUITE_PATH], TestRunnerTags.TEST_FIXTURE: request.fixturename}
            SpanManager.handle_fixture_and_inject_span(HandlerUtils.start_after_each_span, app_info, span_tags,
                request)  
        except Exception as err:
            logger.error("Couldn't start after each span for pytest: {}".format(err))
            pass


    @classmethod
    def finish_after_each_span(cls, request):
        try:
            scope = SpanManager.extract_scope(request)
            span = scope.span
            span.set_tag(TestRunnerTags.TEST_NAME, cls.get_test_method_name(request.node))
            HandlerUtils.finish_after_each_span(scope)
        except Exception as err:
            logger.error("Couldn't finish after each span for pytest: {}".format(err))
            pass


    @classmethod
    def finish_test_span(cls, item):
        try:
            if not hasattr(item, pytest_constants.THUNDRA_TEST_ALREADY_FINISHED):
                setattr(item, pytest_constants.THUNDRA_TEST_ALREADY_FINISHED, True)
                HandlerUtils.finish_test_span()
                cls.clear_test_case_state_for_thundra(item)
        except Exception as err:
            logger.error("Couldn't finish test span for pytest: {}".format(err))
            pass



from foresight.model.terminator import Terminator
from thundra.config.config_provider import ConfigProvider
from thundra.opentracing.tracer import ThundraTracer
from foresight.test_runner_support import TestRunnerSupport
from foresight.environment.environment_info_support import EnvironmentSupport
from foresight.utils.test_wrapper import TestWrapper
from thundra.context.execution_context_manager import ExecutionContextManager
from foresight.test_runner_tags import TestRunnerTags
import foresight.utils.generic_utils as utils
from foresight.model import ThreadExecutorTerminator
import logging

logger = logging.getLogger(__name__)  

class HandlerUtils:

    TEST_BEFORE_ALL_OPERATION_NAME = "beforeAll"
    TEST_AFTER_ALL_OPERATION_NAME = "afterAll"
    TEST_BEFORE_EACH_OPERATION_NAME = "beforeEach"
    TEST_AFTER_EACH_OPERATION_NAME = "afterEach"


    @staticmethod
    def create_span(operation_name, app_info={}, span_tags=None):
        try:
            tracer = ThundraTracer().get_instance()
            execution_context = ExecutionContextManager.get()
            parent_transaction_id = execution_context.transaction_id if execution_context.transaction_id else None
            trace_id = execution_context.trace_id if execution_context.trace_id else utils.create_uuid4()
            transaction_id = parent_transaction_id or utils.create_uuid4()
            scope =  tracer.start_active_span(
                span_id=utils.create_uuid4(),
                operation_name=operation_name,
                trace_id=trace_id,
                transaction_id=transaction_id,
                start_time=utils.current_milli_time(),
                execution_context = execution_context
            )
            span = scope.span
            span.domain_name = app_info.get("applicationDomainName")
            span.class_name = app_info.get("applicationClassName")
            span.service_name = app_info.get("applicationName")
            if span_tags:
                for key in span_tags.keys():
                    span.set_tag(key, span_tags[key])
            return scope
        except Exception as e:
            logger.error("Handler_utils create_span error: {}".format(e))
            pass
        return None

    @staticmethod
    def finish_span(scope, tagName):
        try:
            context = ExecutionContextManager.get()
            current_span = scope.span
            current_span.finish(f_time=utils.current_milli_time())
            scope.close()
            if not context or not context.invocation_data:
                logger.error("Context or context's invocation data couldn't found for finishing trace.")
                return
            invocation_data = context.invocation_data
            if invocation_data:
                duration = current_span.get_duration()
                current_duration = 0
                if tagName in invocation_data["tags"]:
                    current_duration = invocation_data["tags"][tagName]
                duration = duration + current_duration
                invocation_data["tags"][tagName] = duration
        except Exception as e:
            logger.error("Handler_utils finish_span error: {}".format(e))
            pass

    @staticmethod
    def test_setup(executor, api_key=None):
        """Setup foresight environment. already_configured is used for Thundra config is already set or not.

        Args:
            executor (ForesightExecutor): Keeps functions for start and finish thundra plugins
            api_key (str, optional): Thundra api key to send data rest collector. Defaults to None.
        """
        try:
            import thundra 
            already_configured = True if ConfigProvider.configs else False
            thundra._set_thundra_for_test_env(already_configured)
            EnvironmentSupport.init()
            TestWrapper(api_key=api_key, plugin_executor = executor)
            TestRunnerSupport.start_test_run()
        except Exception as e:
            logger.error("Thundra couldn't initialized for test: {}".format(e))
            pass


    @classmethod
    def test_teardown(cls):
        """For now, TestRunnerSupport not support concurrency. Only one test suite info is kept there.
        It should be changed for concurrent python test framework.
        """
        try:
            if (TestRunnerSupport.test_suite_execution_context and 
                not TestRunnerSupport.test_suite_execution_context.completed):
                HandlerUtils.finish_test_suite_span()
            TestRunnerSupport.finish_current_test_run()
            terminator = Terminator()
            terminator.wait(timeout=30)
        except Exception as e:
            logger.error("Handler_utils test_teardown error: {}".format(e))
            pass

    @classmethod
    def start_test_suite_span(cls, test_suite_id, app_info):
        """For now, TestRunnerSupport not support concurrency. Only one test suite info is kept there.
        It should be changed for concurrent python test framework.
        """
        try:
            if not TestRunnerSupport.test_suite_execution_context:
                test_wrapper = TestWrapper.get_instance()
                context = test_wrapper.create_test_suite_execution_context(test_suite_id)
                ExecutionContextManager.set(context)
                test_wrapper.change_app_info(app_info)
                TestRunnerSupport.set_test_suite_application_info(app_info)
                TestRunnerSupport.set_test_suite_execution_context(context)
                test_wrapper.before_test_process(context)
        except Exception as e:
            logger.error("Handler_utils start_test_suite_span error: {}".format(e))
            pass

    @classmethod
    def start_before_all_span(cls, app_info, span_tags= None):
        try:
            return cls.create_span(cls.TEST_BEFORE_ALL_OPERATION_NAME, app_info, span_tags)
        except Exception as e:
            logger.error("Handler_utils start_before_all_span error: {}".format(e))
            pass
        return None

    @classmethod
    def finish_before_all_span(cls, scope):
        try:
            cls.finish_span(scope, TestRunnerTags.TEST_BEFORE_ALL_DURATION)
        except Exception as e:
            logger.error("Handler_utils finish_before_all_span error: {}".format(e))
            pass

    @classmethod
    def start_after_all_span(cls, app_info, span_tags):
        """after all executed after test cases. That is why context should be getting from TestRunnerSupport and
        set into ExecutionContextManager.
        """
        try:
            context = TestRunnerSupport.test_suite_execution_context
            ExecutionContextManager.set(context)
            return cls.create_span(cls.TEST_AFTER_ALL_OPERATION_NAME, app_info, span_tags)
        except Exception as e:
            logger.error("Handler_utils start_after_all_span error: {}".format(e))
            pass
        return None

    @classmethod
    def finish_after_all_span(cls, scope):
        try:
            cls.finish_span(scope, TestRunnerTags.TEST_AFTER_ALL_DURATION)
        except Exception as e:
            logger.error("Handler_utils finish_after_all_span error: {}".format(e))
            pass

    @staticmethod
    def finish_test_suite_span():
        try:
            test_wrapper = TestWrapper.get_instance()
            context = ExecutionContextManager.get()
            context.completed = True
            test_wrapper.after_test_process(context)
            TestRunnerSupport.clear_state()
        except Exception as e:
            logger.error("Handler_utils finish_test_suite_span error: {}".format(e))
            pass

    @classmethod
    def start_test_span(cls, name, test_suite_name, test_case_id, app_info):
        try:
            test_wrapper = TestWrapper.get_instance()
            test_wrapper.change_app_info(app_info)
            current_context = ExecutionContextManager.get()
            parent_transaction_id = current_context.invocation_data.get("transactionId")
            context = test_wrapper.create_test_case_execution_context(name, test_suite_name, test_case_id, app_info, parent_transaction_id)   
            ExecutionContextManager.set(context)
            test_wrapper.before_test_process(context)
        except Exception as e:
            logger.error("Handler_utils start_test_span error: {}".format(e))
            pass

    @classmethod
    def start_before_each_span(cls, app_info, span_tags=None):
        try:
            return cls.create_span(cls.TEST_BEFORE_EACH_OPERATION_NAME, app_info, span_tags)
        except Exception as e:
            logger.error("Handler_utils start_before_each_span error: {}".format(e))
            pass
        return None

    @classmethod
    def finish_before_each_span(cls, scope):
        try:
            cls.finish_span(scope, TestRunnerTags.TEST_BEFORE_EACH_DURATION)
        except Exception as e:
            logger.error("Handler_utils finish_before_each_span error: {}".format(e))
            pass

    @classmethod
    def start_after_each_span(cls, app_info, span_tags=None):
        try:
            return cls.create_span(cls.TEST_AFTER_EACH_OPERATION_NAME, app_info, span_tags)
        except Exception as e:
            logger.error("Handler_utils start_after_each_span error: {}".format(e))
            pass
        return None

    @classmethod
    def finish_after_each_span(cls, scope):
        try:
            cls.finish_span(scope, TestRunnerTags.TEST_AFTER_EACH_DURATION)
        except Exception as e:
            logger.error("Handler_utils finish_after_each_span error: {}".format(e))
            pass

    @staticmethod
    def finish_test_span():
        """Setup TestRunnerSupport for test suite. It's executed after all process has been done 
        for test case such as before_each, after each etc.
        """
        try:
            test_wrapper = TestWrapper.get_instance()
            context = ExecutionContextManager.get()
            test_wrapper.after_test_process(context)
            app_info = TestRunnerSupport.test_suite_application_info
            context = TestRunnerSupport.test_suite_execution_context
            test_wrapper.change_app_info(app_info)
            ExecutionContextManager.set(context)
        except Exception as e:
            logger.error("Handler_utils finish_test_span error: {}".format(e))


from foresight.environment.git.git_helper import GitHelper
from foresight.environment.environment_info import EnvironmentInfo
from foresight.utils.test_runner_utils import TestRunnerUtils
import json, os, logging
from foresight.utils.generic_utils import print_debug_message_to_console


LOGGER = logging.getLogger(__name__)


class GithubEnvironmentInfoProvider:
    ENVIRONMENT = "GitHub"
    GITHUB_REPOSITORY_ENV_VAR_NAME = "GITHUB_REPOSITORY"
    GITHUB_HEAD_REF_ENV_VAR_NAME = "GITHUB_HEAD_REF"
    GITHUB_REF_ENV_VAR_NAME = "GITHUB_REF"
    GITHUB_EVENT_PATH_ENV_VAR_NAME = "GITHUB_EVENT_PATH"
    GITHUB_SHA_ENV_VAR_NAME = "GITHUB_SHA"
    GITHUB_RUN_ID_ENV_VAR_NAME = "GITHUB_RUN_ID"
    INVOCATION_ID_ENV_VAR_NAME = "INVOCATION_ID"
    REFS_HEADS_PREFIX = "refs/heads/"

    @classmethod
    def get_test_run_id(cls, repo_url, commit_hash):
        configured_test_run_id = TestRunnerUtils.get_configured_test_run_id()
        if configured_test_run_id:
            return configured_test_run_id
        run_id = os.getenv(cls.GITHUB_RUN_ID_ENV_VAR_NAME)
        if run_id:
            test_run_key = run_id
            invocation_id = os.getenv(cls.INVOCATION_ID_ENV_VAR_NAME)
            if invocation_id:
                test_run_key = TestRunnerUtils.string_concat_by_underscore(run_id, invocation_id)
            return TestRunnerUtils.get_test_run_id(cls.ENVIRONMENT, repo_url, commit_hash, 
                test_run_key)
        else:
            return TestRunnerUtils.get_default_test_run_id(cls.ENVIRONMENT, repo_url, commit_hash)


    @classmethod
    def build_env_info(cls):
        try:
            github_repo = os.getenv(cls.GITHUB_REPOSITORY_ENV_VAR_NAME)
            repo_url = "https://github.com/" + github_repo + ".git"
            repo_name = GitHelper.extractRepoName(github_repo)
            branch = os.getenv(cls.GITHUB_HEAD_REF_ENV_VAR_NAME)
            commit_hash = os.getenv(cls.GITHUB_SHA_ENV_VAR_NAME)
            commit_message = GitHelper.get_commit_message()
            github_event_path = os.getenv(cls.GITHUB_EVENT_PATH_ENV_VAR_NAME)
            if github_event_path:
                try:
                    if os.path.isfile(github_event_path):
                        with open(github_event_path) as f:
                            try:
                                event_data = json.loads(f.read())
                                from jsonpath_ng import jsonpath, parse
                                jsonpath_expression = parse('$.pull_request.head.sha')
                                match = jsonpath_expression.find(event_data)
                                commit_hash = match[0].value
                            except Exception as err:
                                LOGGER.error("Github Unable to get json data from  GitHub event file " + github_event_path);
                except Exception as err:
                    print_debug_message_to_console("Unable to read GitHub event from file " + github_event_path)
                    LOGGER.error("Github Unable to read GitHub event from file " + github_event_path)
                    pass

            if not branch:
                branch = os.getenv(cls.GITHUB_REF_ENV_VAR_NAME)
                if branch and branch.startswith(cls.REFS_HEADS_PREFIX):
                    branch = branch[len(cls.REFS_HEADS_PREFIX):]

            if not branch:
                branch = GitHelper.get_branch()

            if not commit_hash:
                commit_hash = GitHelper.get_commit_hash()
            
            test_run_id = cls.get_test_run_id(repo_url, commit_hash)
            env_info = EnvironmentInfo(test_run_id, cls.ENVIRONMENT, repo_url, repo_name, branch, commit_hash, commit_message)
            print_debug_message_to_console("Github Environment info: {}".format(env_info.to_json()))
            return env_info
        except Exception as err:
            print_debug_message_to_console("Github Unable to build environment info: {}".format(err))
            LOGGER.error("Github Unable to build environment info: {}".format(err))
            pass
        return None


from thundra.context.execution_context_manager import ExecutionContextManager
from thundra.wrappers.django.django_wrapper import DjangoWrapper, logger
from thundra.wrappers.web_wrapper_utils import process_request_route


class ThundraMiddleware(object):
    def __init__(self, get_response=None):
        super(ThundraMiddleware, self).__init__()
        self.get_response = get_response
        self._wrapper = DjangoWrapper()

    def __call__(self, request):
        setattr(request, '_thundra_wrapped', True)
        # Code to be executed for each request before
        # the view (and later middleware) are called.
        before_done = False
        try:
            self._wrapper.before_request(request)
            before_done = True
        except Exception as e:
            logger.error("Error during the before part of Thundra: {}".format(e))

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.
        if before_done:
            try:
                self._wrapper.after_request(response)
            except Exception as e:
                logger.error("Error during the after part of Thundra: {}".format(e))

        return response

    def process_view(self, request, view_func, view_args, view_kwargs):
        execution_context = ExecutionContextManager.get()
        if request.resolver_match:
            request_host = request.get_host().split(':')[0]
            import sys
            if sys.version_info[0] >= 3:
                process_request_route(execution_context, request.resolver_match.route, request_host)
            else:
                process_request_route(execution_context, request.resolver_match.url_name, request_host)
            
    def process_exception(self, request, exception):
        self._wrapper.process_exception(exception)



from __future__ import absolute_import

from thundra import constants
from thundra.listeners.thundra_span_listener import ThundraSpanListener
from thundra.plugins.invocation import invocation_support

try:
    import builtins
except:
    import __builtin__ as builtins


class SecurityAwareSpanListener(ThundraSpanListener):
    def __init__(self, block=False, whitelist=None, blacklist=None):
        self.block = block
        self.whitelist = whitelist
        self.blacklist = blacklist

    def on_span_started(self, span):
        if not self.is_external_operation(span):
            return

        has_whitelist = isinstance(self.whitelist, list)
        has_blacklist = isinstance(self.blacklist, list)

        if has_blacklist:
            for op in self.blacklist:
                if op.matches(span):
                    self.handle_security_issue(span)
                    return

        if has_whitelist:
            for op in self.whitelist:
                if op.matches(span):
                    return

            self.handle_security_issue(span)

    def on_span_finished(self, span):
        pass

    @staticmethod
    def from_config(config):
        kwargs = {}
        kwargs['block'] = config.get('block')
        whitelist = config.get('whitelist', None)
        blacklist = config.get('blacklist', None)

        if isinstance(whitelist, list):
            kwargs['whitelist'] = list(map(Operation, whitelist))

        if isinstance(blacklist, list):
            kwargs['blacklist'] = list(map(Operation, blacklist))

        return SecurityAwareSpanListener(**kwargs)

    def is_external_operation(self, span):
        return span.get_tag(constants.SpanTags['TOPOLOGY_VERTEX']) == True

    @staticmethod
    def should_raise_exceptions():
        return True

    def handle_security_issue(self, span):
        if self.block:
            err = SecurityError()
            span.set_tag(constants.SecurityTags['VIOLATED'], True)
            span.set_tag(constants.SecurityTags['BLOCKED'], True)
            span.set_error_to_tag(err)
            invocation_support.set_agent_tag(constants.SecurityTags['VIOLATED'], True)
            invocation_support.set_agent_tag(constants.SecurityTags['BLOCKED'], True)
            raise err
        else:
            span.set_tag(constants.SecurityTags['VIOLATED'], True)
            invocation_support.set_agent_tag(constants.SecurityTags['VIOLATED'], True)


class SecurityError(Exception):
    def __init__(self, msg='Operation was blocked due to security configuration'):
        super(Exception, self).__init__(msg)


class Operation:
    def __init__(self, config):
        self.class_name = config.get('className')
        self.operation_name = config.get('operationName')
        self.tags = config.get('tags')

    def matches(self, span):
        matched = self.class_name == span.class_name or self.class_name == '*'

        if matched and isinstance(self.operation_name, list):
            matched = span.operation_name in self.operation_name or '*' in self.operation_name

        if matched and self.tags:
            for k, v in self.tags.items():
                if isinstance(v, list):
                    if not (span.get_tag(k) in v or '*' in v):
                        matched = False
                        break
                elif span.get_tag(k) != v:
                    matched = False
                    break
        return matched



import os

from thundra.config import config_names
from thundra.config.config_metadata import CONFIG_METADATA


class ConfigProvider:
    configs = {}

    @staticmethod
    def __init__(options=None):
        ConfigProvider.clear()
        if options is not None:
            config_options = options.get('config', {})
            for opt in config_options:
                if opt.lower() == config_names.THUNDRA_APIKEY:
                    ConfigProvider.configs[config_names.THUNDRA_APIKEY] = config_options.get(opt)
                ConfigProvider.traverse_config_object(config_options.get(opt), opt)
        ConfigProvider.initialize_config_from_environment_variables()
        ConfigProvider.add_non_lambda_aliases()

    @staticmethod
    def initialize_config_from_environment_variables():
        env_variables = os.environ

        for var_name in env_variables:
            if var_name.upper().startswith("THUNDRA_"):
                env_var_name = ConfigProvider.env_var_to_config_name(var_name)
                val = env_variables.get(var_name).strip()
                env_var_type = ConfigProvider.get_config_type(env_var_name)
                ConfigProvider.configs[env_var_name] = ConfigProvider.parse(val, env_var_type)

    @staticmethod
    def traverse_config_object(obj, path):
        if not isinstance(obj, dict):
            if not path.startswith('thundra.agent.'):
                path = 'thundra.agent.' + path
            path = path.lower()
            prop_type = ConfigProvider.get_config_type(path)
            ConfigProvider.configs[path] = ConfigProvider.parse(obj, prop_type)
        else:
            for prop_name in obj:
                prop_val = obj.get(prop_name)
                prop_path = path + '.' + prop_name
                ConfigProvider.traverse_config_object(prop_val, prop_path)

    @staticmethod
    def add_non_lambda_aliases():
        alias_configs = {}
        for config_name in ConfigProvider.configs:
            if config_name.startswith('thundra.agent.lambda'):
                value = ConfigProvider.configs[config_name]
                alias_configs[config_name.replace('thundra.agent.lambda', 'thundra.agent', 1)] = value
        ConfigProvider.configs.update(alias_configs)

    @staticmethod
    def get(key, default_value=None):
        value = ConfigProvider.configs.get(key)
        if value is not None:
            return value
        if default_value is not None:
            return default_value
        if CONFIG_METADATA.get(key):
            return CONFIG_METADATA[key].get('defaultValue')
        return None

    @staticmethod
    def get_config_type(config_name):
        config_metadata = CONFIG_METADATA.get(config_name)
        if config_metadata:
            return config_metadata.get('type')
        else:
            if config_name.startswith('thundra.agent.lambda.'):
                config_name = config_name.replace('thundra.agent.lambda.', 'thundra.agent.', 1)
                config_metadata = CONFIG_METADATA.get(config_name)
                if config_metadata:
                    return config_metadata.get('type')
        return None

    @staticmethod
    def parse(value, var_type):
        if var_type == 'string':
            return value
        if var_type == 'int':
            return ConfigProvider.convert_to_int(value)
        if var_type == 'boolean':
            return ConfigProvider.convert_to_bool(value)
        return ConfigProvider.str_to_proper_type(value)

    @staticmethod
    def str2bool(val):
        if type(val) == bool:
            return val
        if isinstance(val, str):
            if val.lower() in ("yes", "true", "t", "1"):
                return True
            elif val.lower() in ("no", "false", "f", "0"):
                return False
        raise ValueError

    @staticmethod
    def str_to_proper_type(val):
        try:
            result = ConfigProvider.str2bool(val)
        except ValueError:
            try:
                result = int(val)
            except ValueError:
                try:
                    result = float(val)
                except ValueError:
                    result = val.strip('"')

        return result

    @staticmethod
    def convert_to_bool(value, default=False):
        if type(value) == bool:
            return value
        try:
            return ConfigProvider.str2bool(value)
        except ValueError:
            return default

    @staticmethod
    def convert_to_int(value, default=0):
        try:
            return int(value)
        except (ValueError, TypeError):
            return default

    @staticmethod
    def config_name_to_env_var(config_name):
        return config_name.upper().replace('.', '_')

    @staticmethod
    def env_var_to_config_name(env_var_name):
        return env_var_name.lower().replace('_', '.')

    @staticmethod
    def clear():
        ConfigProvider.configs.clear()

    @staticmethod
    def set(key, value):
        ConfigProvider.configs[key] = ConfigProvider.parse(value, ConfigProvider.get_config_type(key))


ConfigProvider.__init__()



from __future__ import absolute_import
import logging
import os
import re
import uuid

from thundra import constants
from thundra.compat import urlparse
from opentracing.scope_managers import ThreadLocalScopeManager

logger = logging.getLogger(__name__)


def generate_id():
    return str(uuid.uuid4())


def generate_id_from(value):
    return str(uuid.uuid5(uuid.NAMESPACE_OID, value))


def get_env_variable(key, default=None):
    return os.environ.get(key, default)


def get_aws_function_name(arn):
    return get_arn_part(arn, 6)


def get_aws_region_from_arn(arn):
    return get_arn_part(arn, 3)


def get_aws_account_no(arn):
    return get_arn_part(arn, 4)


def get_arn_part(arn, index):
    # ARN format: arn:aws:lambda:{region}:{account_no}:function:{function_name}
    try:
        return arn.split(":")[index]
    except:
        return ""


def get_aws_lambda_function_memory_size():
    return os.environ.get(constants.AWS_LAMBDA_FUNCTION_MEMORY_SIZE)


def sam_local_debugging():
    return os.environ.get(constants.AWS_SAM_LOCAL) == 'true'


#### memory ####
def process_memory_usage():
    try:
        with open('/proc/self/status', 'r') as procfile:
            mems = {}
            for line in procfile:
                fields = line.split(':')
                try:
                    mem_key = fields[0]
                    mem_val = (fields[1].split())[0]
                    mems[mem_key] = mem_val
                except IndexError:
                    continue

            size_from_env_var = get_aws_lambda_function_memory_size()
            if not size_from_env_var:
                size = int(mems['VmSize'])
                size_in_bytes = int(size * 1024)
            else:
                size_in_bytes = int(float(size_from_env_var) * 1048576)

            used_mem_in_kb = int(mems['VmRSS']) + int(mems['VmSwap'])
            used_mem_in_bytes = used_mem_in_kb * 1024

            return size_in_bytes, used_mem_in_bytes
    except Exception as e:
        logger.error('ERROR: {}'.format(e))
        return 0, 0


##### cpu #####
def process_cpu_usage():
    try:
        with open('/proc/self/stat', 'r') as procfile:
            process_cpu_usages = procfile.readline()
            # get utime from /proc/<pid>/stat, 14 item
            u_time = process_cpu_usages.split(' ')[13]
            # get stime from proc/<pid>/stat, 15 item
            s_time = process_cpu_usages.split(' ')[14]
            # count total process used time
            process_cpu_used = int(u_time) + int(s_time)
            return (float(process_cpu_used))
    except Exception as e:
        logger.error('ERROR: {}'.format(e))
        return 0


def system_cpu_usage():
    try:
        with open('/proc/stat', 'r') as procfile:
            system_cpu_usages = procfile.readline()
            system_cpu_used = 0
            system_cpu_total = 0
            count = 0
            for usage in system_cpu_usages.split(' ')[2:]:
                if count == 5:
                    break
                elif count != 3 and count != 4:
                    system_cpu_used += int(usage)
                system_cpu_total += int(usage)
                count += 1
            return float(system_cpu_total), float(system_cpu_used)
    except Exception as e:
        logger.error('ERROR: {}'.format(e))
        return 0, 0


#####################################################################
###
#####################################################################


class _Singleton(type):
    """ A metaclass that creates a Singleton base class when called. """
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

class Singleton(_Singleton('SingletonMeta', (object,), {})): pass

def get_all_env_variables():
    return os.environ


def get_module_name(module):
    return module.__name__


def string_to_list(target, indicator):
    return target.split(indicator)


def str2bool(val):
    if val is not None:
        if val.lower() in ("yes", "true", "t", "1"):
            return True
        elif val.lower() in ("no", "false", "f", "0"):
            return False
    raise ValueError


def process_trace_def_var(value):
    value = value.strip().split('[')
    path = value[0].split('.')
    trace_args = {}

    function_prefix = path[-1][:-1] if path[-1] != '*' else ''
    module_path = ".".join(path[:-1])
    trace_string = value[1].strip(']').split(',')
    for arg in trace_string:
        arg = arg.split('=')
        try:
            trace_args[arg[0]] = arg[1]
        except:
            pass

    return module_path, function_prefix, trace_args


def get_allowed_functions(module):
    allowed_functions = []
    for prop in vars(module):
        # TO DO: differentiate functions
        allowed_functions.append(str(prop))
    return allowed_functions


def is_excluded_url(url):
    host = urlparse(url).netloc

    for method in EXCLUDE_EXCEPTION_URLS:
        for exclude_exception_url in EXCLUDE_EXCEPTION_URLS[method]:
            if method(host, exclude_exception_url):
                return False

    for method in EXCLUDED_URLS:
        for excluded_url in EXCLUDED_URLS[method]:
            if method(host, excluded_url):
                return True
    return False


def get_default_timeout_margin():
    region = get_env_variable(constants.AWS_REGION, default='')
    size_from_env_var = get_aws_lambda_function_memory_size()
    memory = -1
    if size_from_env_var:
        memory = int(float(size_from_env_var))

    timeout_margin = 1000

    if region == 'us-west-2':
        timeout_margin = 200
    elif region.startswith('us-west-'):
        timeout_margin = 400
    elif region.startswith('us-') or region.startswith('ca-'):
        timeout_margin = 600
    elif region.startswith('sa-'):
        timeout_margin = 800

    normalized_timeout_margin = int((384.0 / memory) * timeout_margin)
    return max(timeout_margin, normalized_timeout_margin)


def parse_x_ray_trace_info():
    xray_trace_header = os.environ.get("_X_AMZN_TRACE_ID")
    xray_info = {"trace_id": None, "segment_id": None}
    if xray_trace_header:
        for trace_header_part in xray_trace_header.split(";"):
            trace_info = trace_header_part.split("=")
            if len(trace_info) == 2 and trace_info[0] == "Root":
                xray_info["trace_id"] = trace_info[1]
            elif len(trace_info) == 2 and trace_info[0] == "Parent":
                xray_info["segment_id"] = trace_info[1]

    return xray_info


def get_nearest_collector():
    region = get_env_variable(constants.AWS_REGION)

    if region:
        return '{}.collector.thundra.io'.format(region)

    return "collector.thundra.io"


def get_compiled_operation_type_patterns():
    compiled = []
    for pattern in constants.OperationTypeMappings["patterns"]:
        compiled.append(re.compile(pattern))
    return compiled


def extract_api_gw_resource_name(event):
    try:
        resource_path = None
        if 'resource' in event:
            resource_path = event['resource']
        else:
            resource_path = event['requestContext']['http']['path']
            stage_prefix = '/' + event['requestContext']['stage']
            if resource_path.startswith(stage_prefix):
                resource_path = resource_path[len(stage_prefix):]

        return resource_path
    except:
        return None


def get_normalized_path(url_path, path_depth):
    path_seperator_count = 0
    normalized_path = ''
    prev_c = ''
    for c in url_path:
        if c == '/' and prev_c != '/':
            path_seperator_count += 1

        if path_seperator_count > path_depth:
            break

        normalized_path += c
        prev_c = c
    return normalized_path


def parse_http_url(url, url_path_depth):
    url_dict = {
        'path': '',
        'query': '',
        'host': '',
        'url': url
    }
    try:
        parsed_url = urlparse(url)
        url_dict['path'] = parsed_url.path
        url_dict['query'] = parsed_url.query
        url_dict['host'] = parsed_url.netloc

        normalized_path = get_normalized_path(parsed_url.path, url_path_depth)
        url_dict['operation_name'] = parsed_url.hostname + normalized_path

        url_dict['url'] = parsed_url.hostname + parsed_url.path
    except Exception:
        pass
    return url_dict


def arrange_scope_manager(scope_manager):
    if scope_manager is None:
        try:
            import sys
            if (sys.version_info[0] > 3) or (sys.version_info[0] == 3 and sys.version_info[1] > 7) or (sys.version_info[0] == 3 and sys.version_info[1] == 7 and sys.version_info[2] != 0):
                from opentracing.scope_managers.contextvars import ContextVarsScopeManager
                scope_manager = ContextVarsScopeManager()
            else:
                scope_manager = ThreadLocalScopeManager()
        except Exception:
            scope_manager = ThreadLocalScopeManager()
    return scope_manager


# Excluded url's
EXCLUDED_URLS = {
    str.endswith: [
        'thundra.io',
    ],
    str.__contains__: [
        'amazonaws.com',
        'accounts.google.com',
    ],
}

# Exclude exception urls 
EXCLUDE_EXCEPTION_URLS = {
    str.__contains__: [
        '.execute-api.',
        '.elb.'
    ]
}



import wrapt

from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider


def _wrapper(wrapped, instance, args, kwargs):
    try:
        from thundra.integrations.aiohttp.client import ThundraTraceConfig
        trace_configs = kwargs.get('trace_configs', [])
        trace_configs.append(ThundraTraceConfig())
        kwargs['trace_configs'] = trace_configs
    except:
        pass
    wrapped(*args, **kwargs)


def patch():
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_DISABLE):
        try:
            import aiohttp
            wrapt.wrap_function_wrapper(
                'aiohttp',
                'ClientSession.__init__',
                _wrapper
            )
        except ImportError:
            pass



from thundra import constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider

from thundra.wrappers import wrapper_utils, web_wrapper_utils
import logging

Logger = logging.getLogger(__name__)


def start_trace(plugin_context, execution_context, tracer):
    request = execution_context.platform_data['request']
    request_route_path = str(request.url_rule) if request.url_rule else None
    '''
        request data read as stream. This function cache the read data and serve from cache.
        According to function comment in Flask, content length should be checked before read!!! 
    '''
    _request = {
        'method': request.method,
        'host': request.host.split(':')[0],
        'query_params': request.query_string,
        'headers': request.headers,
        'body': None,
        'path': request.path
    }

    web_wrapper_utils.start_trace(execution_context, tracer, 'Flask', 'API', _request, request_route_path)


def finish_trace(execution_context):
    root_span = execution_context.root_span
    """
        Getting request data into start trace occurs unexpected bugs into application process.
        After whole process finish for request data into appşication flow, getting request data and
        set the request_body for root_span as tag.
    """
    try:
        _request = execution_context.platform_data['request']
        req_data = None
        if _request and not ConfigProvider.get(config_names.THUNDRA_TRACE_REQUEST_SKIP, False):
            cl = _request.content_length
            if cl == None or cl <= constants.THUNDRA_MAX_STREAM_REQUEST_BODY:
                req_data = _request.get_data()
            else:
                req_data = None
            root_span.set_tag(constants.HttpTags['BODY'], req_data)
    except Exception as e:
        Logger.error("Error occured whilst setting request body to root span tag: {}".format(e))
        pass
    if execution_context.response:
        status_code = get_response_status(execution_context)
        if status_code:
            root_span.set_tag(constants.HttpTags['HTTP_STATUS'], status_code)
        if execution_context.trigger_operation_name and execution_context.response and hasattr(
                execution_context.response, 'headers'):
            execution_context.response.headers[
                constants.TRIGGER_RESOURCE_NAME_KEY] = execution_context.trigger_operation_name
    web_wrapper_utils.finish_trace(execution_context)


def start_invocation(plugin_context, execution_context):
    execution_context.invocation_data = wrapper_utils.create_invocation_data(plugin_context, execution_context)


def finish_invocation(execution_context):
    wrapper_utils.finish_invocation(execution_context)

    # Set response status code
    wrapper_utils.set_response_status(execution_context, get_response_status(execution_context))


def get_response_status(execution_context):
    try:
        status_code = execution_context.response.status_code
    except:
        return None
    return status_code



import logging
import time
import traceback

from thundra import constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.rdb_base import RdbBaseIntegration
from thundra.opentracing.tracer import ThundraTracer

try:
    from sqlalchemy.event import listen
    from sqlalchemy.engine.interfaces import ExecutionContext
except:
    pass

logger = logging.getLogger(__name__)


class SqlAlchemyIntegration(RdbBaseIntegration):
    CLASS_TYPE = 'sqlalchemy'

    def __init__(self, engine):
        listen(engine, 'before_cursor_execute', self._before_cursor_execute)
        listen(engine, 'after_cursor_execute', self._after_cursor_execute)
        try:
            listen(engine, 'handle_error', self._handle_error)
        except:
            listen(engine, 'dbapi_error', self._db_api_error)

    def get_operation_name(self, conn):
        return self.get_db_config(conn).get('operation_name', '')

    def get_db_config(self, conn):
        params = {}
        try:
            db_name = conn.engine.name
            if 'postgres' in db_name or 'psycopg2' in db_name:
                params['db_type'] = 'postgresql'
            elif 'sqlite' in db_name:
                params['db_type'] = 'sqlite'
            elif 'mysql' in db_name:
                params['db_type'] = 'mysql'
            else:
                params['db_type'] = db_name

            _url = conn.engine.url
            if _url.host:
                params['host'] = _url.host
                params['operation_name'] = _url.host

            if _url.database:
                params['database'] = _url.database
                params['operation_name'] = _url.database
        except:
            pass

        return params

    def before_call(self, scope, conn, statement):
        db_config = self.get_db_config(conn)

        scope.span.class_name = constants.ClassNames.get(db_config.get('db_type', '').upper())
        scope.span.domain_name = constants.DomainNames['DB']

        try:
            operation = statement.split()[0].strip("\"").lower()
        except:
            operation = ""

        tags = {
            constants.SpanTags['OPERATION_TYPE']: SqlAlchemyIntegration._OPERATION_TO_TYPE.get(operation, ''),
            constants.SpanTags['DB_INSTANCE']: db_config.get('database', ''),
            constants.SpanTags['DB_HOST']: db_config.get('host', ''),
            constants.SpanTags['DB_TYPE']: db_config.get('db_type', ''),
            constants.SpanTags['DB_STATEMENT_TYPE']: operation.upper(),
            constants.SpanTags['TOPOLOGY_VERTEX']: True
        }

        if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_RDB_STATEMENT_MASK):
            tags[constants.DBTags['DB_STATEMENT']] = statement

        scope.span.tags = tags

    def after_call(self, scope, conn, statement):
        pass

    def _before_cursor_execute(self, conn, cursor, statement, parameters, context, executemany):
        if not context:
            return
        tracer = ThundraTracer.get_instance()

        if not tracer.get_active_span():
            return
        scope = tracer.start_active_span(operation_name=self.get_operation_name(conn), finish_on_close=False)

        context.scope = scope

        # Inject before span tags
        try:
            self.before_call(scope, conn, statement)
        except Exception as instrumentation_exception:
            e = {
                'type': str(type(instrumentation_exception)),
                'message': str(instrumentation_exception),
                'traceback': traceback.format_exc(),
                'time': time.time()
            }
            scope.span.set_tag('instrumentation_error', e)

        try:
            # Inform span that initialization completed
            scope.span.on_started()
        except Exception as e:
            logger.error(e)

    def _after_cursor_execute(self, conn, cursor, statement, parameters, context, executemany):
        try:
            scope = context.scope
        except:
            return
        # Inject after span tags 
        try:
            self.after_call(scope, conn, statement)
        except Exception as instrumentation_exception:
            e = {
                'type': str(type(instrumentation_exception)),
                'message': str(instrumentation_exception),
                'traceback': traceback.format_exc(),
                'time': time.time()
            }
            scope.span.set_tag('instrumentation_error', e)

        try:
            scope.span.finish()
        except Exception as e:
            logger.error(e)

        scope.close()

    def _set_exception(self, conn, statement, context, exception):
        try:
            scope = context.scope
        except:
            return

        try:
            scope.span.set_error_to_tag(exception)
        finally:
            scope.span.finish()
            scope.close()

    def _db_api_error(self, conn, cursor, statement, parameters, context, exception):
        self._set_exception(conn, statement, context, exception)

    def _handle_error(self, exception_context):
        self._set_exception(exception_context.connection, exception_context.statement,
                            exception_context.execution_context, exception_context.original_exception)



import logging
import time
from threading import Lock

import opentracing

logger = logging.getLogger(__name__)


class ThundraSpan(opentracing.Span):

    def __init__(self,
                 tracer,
                 operation_name=None,
                 class_name=None,
                 domain_name=None,
                 context=None,
                 tags=None,
                 start_time=None,
                 span_order=-1,
                 execution_context=None):
        super(ThundraSpan, self).__init__(tracer, context)
        self._tracer = tracer
        self._context = context
        self._lock = Lock()
        self.operation_name = operation_name if operation_name is not None else ""
        self.class_name = class_name if class_name is not None else ""
        self.domain_name = domain_name if domain_name is not None else ""
        self.start_time = start_time or int(time.time() * 1000)
        self.finish_time = 0
        self.span_order = span_order
        self.tags = tags if tags is not None else {}
        self.logs = []
        self._execution_context = execution_context

    @property
    def context(self):
        return self._context

    @property
    def execution_context(self):
        return self._execution_context

    @execution_context.setter
    def execution_context(self, execution_context):
        self._execution_context = execution_context

    @property
    def trace_id(self):
        return self._context.trace_id

    @property
    def transaction_id(self):
        return self._context.transaction_id

    @property
    def span_id(self):
        return self._context.span_id

    def set_operation_name(self, operation_name):
        with self._lock:
            self.operation_name = operation_name
        return super(ThundraSpan, self).set_operation_name(operation_name)

    def set_tag(self, key, value):
        with self._lock:
            if self.tags is None:
                self.tags = {}
            self.tags[key] = value
        return super(ThundraSpan, self).set_tag(key, value)

    def get_tag(self, key):
        if self.tags is not None:
            return self.tags.get(key)
        return None

    def finish(self, f_time=None):
        with self._lock:
            self.finish_time = int(time.time() * 1000) if f_time is None else f_time
        self.on_finished()

    def on_finished(self):
        span_listeners = self._tracer.get_span_listeners()
        for sl in span_listeners:
            try:
                sl.on_span_finished(self)
            except Exception as e:
                if not sl.should_raise_exceptions():
                    logger.error(("error while calling"
                                  " on_finished of %s: %s"), type(sl), e)
                else:
                    raise e

    def on_started(self):
        span_listeners = self._tracer.get_span_listeners()
        for sl in span_listeners:
            try:
                sl.on_span_started(self)
            except Exception as e:
                if not sl.should_raise_exceptions():
                    logger.error(("error while calling"
                                  " on_started of %s: %s"), type(sl), e)
                else:
                    raise e

    def log_kv(self, key_values, timestamp=None):
        with self._lock:
            log = key_values
            log['timestamp'] = timestamp
            self.logs.append(log)
        return super(ThundraSpan, self).log_kv(key_values, timestamp)

    def set_baggage_item(self, key, value):
        new_context = self.context.context_with_baggage_item(key, value)
        with self._lock:
            self._context = new_context
        return self

    def get_baggage_item(self, key):
        with self._lock:
            return self.context.baggage.get(key)

    def set_error_to_tag(self, err):
        error_type = type(err)
        self.set_tag('error', True)
        self.set_tag('error.kind', error_type.__name__)
        self.set_tag('error.message', str(err))

    def get_duration(self):
        if self.finish_time == 0:
            return int(time.time() * 1000) - self.start_time
        else:
            return self.finish_time - self.start_time

    def erroneous(self):
        return self.tags is not None and 'error' in self.tags



from foresight.environment.environment_info_support import EnvironmentSupport
from foresight.utils.test_wrapper import TestWrapper
import logging

logger = logging.getLogger(__name__)

def start_trace(plugin_context, execution_context, tracer):
    """Start trace plugin. It is executed by hook in BaseWrapper

    Args:
        plugin_context (PluginContext): Stores thundra plugins' contexts
        execution_context (TestCaseExecutionContext | TestSuiteExecutionContext): Stores Execution context info
        tracer (ThundraTracer): Thundra scope tracer
    """
    try:
        test_wrapper = TestWrapper.get_instance()
        test_wrapper.start_trace(execution_context, tracer)
        
        root_span = execution_context.root_span

        root_span.tags.update(execution_context.get_additional_start_tags())

        EnvironmentSupport.set_span_tags(root_span)
    except Exception as err:
        logger.error("Foresight executor start trace error {}".format(err))
        pass

def finish_trace(execution_context):
    """Finish trace plugin. It is executed by hook in BaseWrapper

    Args:
        execution_context (TestCaseExecutionContext | TestSuiteExecutionContext): Stores Execution context info
    """
    try:
        test_wrapper = TestWrapper.get_instance()
        test_wrapper.finish_trace(execution_context)

        root_span = execution_context

        root_span.tags.update(execution_context.get_additional_finish_tags())
    except Exception as err:
        logger.error("Foresight executor finish trace error: {}".format(err))
        pass

def start_invocation(plugin_context, execution_context):
    """Start invocation for thundra invocation plugin.It is executed by hook in BaseWrapper

    Args:
        plugin_context (PluginContext): Stores thundra plugins' contexts
        execution_context (TestCaseExecutionContext | TestSuiteExecutionContext): Stores Execution context info
    """
    try:
        test_wrapper = TestWrapper.get_instance()
        test_wrapper.start_invocation(execution_context)
        
        execution_context.invocation_data["tags"].update(execution_context.get_additional_start_tags())
    except Exception as err:
        logger.error("Foresight executor start incovation error: {}".format(err))
        pass

def finish_invocation(execution_context):    
    """Finish invocation for thundra invocation plugin.It is executed by hook in BaseWrapper

    Args:
        execution_context (TestCaseExecutionContext | TestSuiteExecutionContext): Stores Execution context info
    """
    try:
        invocation_data = execution_context.invocation_data

        test_wrapper = TestWrapper.get_instance()

        invocation_data["tags"].update(execution_context.get_additional_finish_tags())

        error = execution_context.error
        if error:
            invocation_data['tags']['error.type'] = error.get('type')
            invocation_data['tags']['error.message'] = error.get('message')
            invocation_data['tags']['error.stack'] = error.get('traceback', None)

        EnvironmentSupport.set_invocation_tags(invocation_data)  
        test_wrapper.finish_invocation(execution_context)
    except Exception as err:
        logger.error("Foresight executor finish invocation error: {}".format(err))
        pass


import wrapt

from thundra import utils, constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.wrappers.flask.middleware import ThundraMiddleware


def _wrapper(wrapped, instance, args, kwargs):
    response = wrapped(*args, **kwargs)
    try:
        thundra_middleware = ThundraMiddleware()
        thundra_middleware.set_app(instance)
    except:
        pass
    return response


def patch():
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_FLASK_DISABLE) and (
            not utils.get_env_variable(constants.AWS_LAMBDA_FUNCTION_NAME)):
        wrapt.wrap_function_wrapper(
            'flask',
            'Flask.__init__',
            _wrapper
        )



class TestRunnerTags:
    TEST_RUN_ID = "test.run.id"
    TEST_RUN_TASK_ID = "test.run.task.id"
    TEST_SUITE_TRANSACTION_ID = "test.suite.transaction.id"
    TEST_NAME = "test.name"
    TEST_SUITE = "test.suite"
    TEST_FIXTURE = "test.fixture"
    TEST_STATUS = "test.status"
    TEST_SUITE_TOTAL_COUNT = "test.suite.total.count"
    TEST_SUITE_SUCCESSFUL_COUNT = "test.suite.successful.count"
    TEST_SUITE_FAILED_COUNT = "test.suite.failed.count"
    TEST_SUITE_ABORTED_COUNT = "test.suite.aborted.count"
    TEST_SUITE_SKIPPED_COUNT = "test.suite.skipped.count"
    TEST_METHOD = "test.method"
    TEST_CLASS = "test.class"
    TEST_TIMEOUT = "test.timeout"
    TEST_BEFORE_EACH_DURATION = "test.before.each.duration"
    TEST_AFTER_EACH_DURATION = "test.after.each.duration"
    TEST_BEFORE_ALL_DURATION = "test.before.all.duration"
    TEST_AFTER_ALL_DURATION = "test.after.all.duration"
    TEST_ENVIRONMENT = "test.env"
    SOURCE_CODE_REPO_URL = "src.repo.url"
    SOURCE_CODE_REPO_NAME = "src.repo.name"
    SOURCE_CODE_BRANCH = "src.branch"
    SOURCE_CODE_COMMIT_HASH = "src.commit.hash"
    SOURCE_CODE_COMMIT_MESSAGE = "src.commit.message"


import logging
import uuid

from thundra.context.execution_context_manager import ExecutionContextManager


class ThundraLogHandler(logging.Handler):

    def __init__(self):
        logging.Handler.__init__(self)

    def emit(self, record):
        formatted_message = self.format(record)
        execution_context = ExecutionContextManager.get()
        if execution_context and execution_context.capture_log:
            log = {
                'id': str(uuid.uuid4()),
                'spanId': execution_context.span_id if execution_context is not None else '',
                'logMessage': formatted_message,
                'logContextName': record.name,
                'logTimestamp': int(record.created * 1000),
                'logLevel': record.levelname,
                'logLevelCode': int(record.levelno / 10)
            }
            execution_context.logs.append(log)


logging.ThundraLogHandler = ThundraLogHandler



import inspect
import sys
from functools import wraps
from threading import Lock

import jsonpickle
from opentracing import Scope

from thundra import constants
from thundra.opentracing.tracer import ThundraTracer
from thundra.plugins.log.thundra_logger import debug_logger
from thundra.serializable import Serializable
from pympler import asizeof
import threading
trace_local = threading.local()

DATA_LIMIT = 8 * 1024 # 8KB
DATA_LIMIT_RETURN_STR = "[THUNDRA] Data is over {} KB".format(DATA_LIMIT / 1024)
METHOD_LINES_LIMIT = 100
SERIALIZATION_DEPTH=1

def __get_traceable_from_back_frame(frame):
    _back_frame = frame.f_back
    if _back_frame and 'self' in _back_frame.f_locals:
        _self = _back_frame.f_locals['self']
        if isinstance(_self, Traceable):
            return _self
    return None


def __get_funcname_from_back_frame(frame):
    _back_frame = frame.f_back
    if _back_frame and 'original_func' in _back_frame.f_locals:
        _func = _back_frame.f_locals['original_func']
        return _func.__name__
    return None


def __get_scope_from_back_frame(frame):
    _back_frame = frame.f_back
    if _back_frame and 'scope' in _back_frame.f_locals:
        _scope = _back_frame.f_locals['scope']
        if isinstance(_scope, Scope):
            return _scope
    return None


def trace_lines(frame, event, arg):
    if event != 'line':
        return
    _line_no = frame.f_lineno

    _trace_local_variables_ = False
    _trace_lines_with_source = False
    _scope = __get_scope_from_back_frame(frame)

    if not _scope or not _scope.span:
        return

    _traceable = __get_traceable_from_back_frame(frame)
    if _traceable == None:
        return
    _trace_local_variables_ = _traceable._trace_local_variables
    _trace_lines_with_source = _traceable._trace_lines_with_source

    method_lines_list = _scope.span.get_tag(constants.LineByLineTracingTags['lines'])
    if method_lines_list != None and len(method_lines_list) >= METHOD_LINES_LIMIT:
        return
    _local_vars = []
    global DATA_LIMIT, DATA_LIMIT_RETURN_STR, SERIALIZATION_DEPTH
    if _trace_local_variables_:
        pickler = jsonpickle.pickler.Pickler(max_depth=SERIALIZATION_DEPTH)
        for l in frame.f_locals:
            _local_var_value = frame.f_locals[l]
            _local_var_type = type(_local_var_value).__name__
            if asizeof.asizeof(_local_var_value, limit=SERIALIZATION_DEPTH) > DATA_LIMIT:
                _local_var_value = DATA_LIMIT_RETURN_STR
            else:
                try:
                    _local_var_value = pickler.flatten(_local_var_value, reset=True)
                except Exception as e:
                    _local_var_value = '<not-json-serializable-object>'
            _local_var = {
                'name': l,
                'value': _local_var_value,
                'type': _local_var_type
            }
            _local_vars.append(_local_var)

    method_line = {
        'line': _line_no,
        'localVars': _local_vars
    }
    if not method_lines_list:
        method_lines_list = []
    method_lines_list.append(method_line)
    _scope.span.set_tag(constants.LineByLineTracingTags['lines'], method_lines_list)
    

def trace_calls(frame, event, arg):
    if not trace_local.trace_call_active:
        return

    trace_local.trace_call_active = False

    # First check whether current call is wrapped
    wrapped_by_thundra = frame.f_back and frame.f_back.f_code.co_name == '___thundra_trace___'
    if not wrapped_by_thundra:
        return

    # Note that "is Thundra wrapped check" is applied before "is event call" check.
    # Because most of the time "is Thundra wrapped check" is false but "is event call" is true.
    # And since most of the time 'trace_calls' will return 'None',
    # it is better to detect false case with less comparison.

    # Only handle 'call' events
    if event != 'call':
        return

    _func_name = frame.f_code.co_name
    # Ignore
    # - 'write()' calls from print statements
    # - Thundra trace decorator calls
    ignored = _func_name == 'write' or _func_name == '___thundra_trace___'
    if ignored:
        return

    _traceable = None
    try:
        _traceable = __get_traceable_from_back_frame(frame)
    except ValueError:
        pass
    if not _traceable or not _traceable.trace_line_by_line:
        return

    orig_func_name = __get_funcname_from_back_frame(frame)
    if _func_name == orig_func_name:
        return trace_lines


# To keep track of the active line-by-line traced Tracable count
_lock = Lock()
_line_traced_count = 0

class Traceable:

    def __init__(self,
                 trace_args=None, trace_return_value=None, trace_error=True,
                 trace_line_by_line=False, trace_lines_with_source=None, trace_local_variables=None):
        self._trace_args = trace_args
        if trace_args is None:
            self._trace_args = False

        self._trace_return_value = trace_return_value
        if trace_return_value is None:
            self._trace_return_value = False

        self._trace_lines_with_source = trace_lines_with_source
        if trace_lines_with_source is None:
            self._trace_lines_with_source = False

        self._trace_local_variables = trace_local_variables
        if trace_local_variables is None:
            self._trace_local_variables = False

        self._trace_error = trace_error
        self._trace_line_by_line = trace_line_by_line

        if trace_line_by_line:
            if trace_lines_with_source is None:
                self._trace_lines_with_source = True
            if trace_return_value is None:
                self._trace_return_value = True
            if trace_local_variables is None:
                self._trace_local_variables = True
            if trace_args is None:
                self._trace_args = True

        self._tracing = False
        self._tracer = ThundraTracer.get_instance()

    @property
    def tracer(self):
        return self._tracer

    @property
    def trace_args(self):
        return self._trace_args

    @property
    def trace_return_value(self):
        return self._trace_return_value

    @property
    def trace_error(self):
        return self._trace_error

    @property
    def trace_line_by_line(self):
        return self._trace_line_by_line

    @property
    def trace_lines_with_source(self):
        return self._trace_lines_with_source

    @property
    def trace_local_variables(self):
        return self._trace_local_variables

    def __is_serializable__(self, value):
        return value is None or isinstance(value, (str, int, float, bool))

    def __serialize_value__(self, value):
        global DATA_LIMIT, DATA_LIMIT_RETURN_STR, SERIALIZATION_DEPTH
        value_size = asizeof.asizeof(value, limit=SERIALIZATION_DEPTH)
        if value_size > DATA_LIMIT:
            return DATA_LIMIT_RETURN_STR, type(value).__name__
        if self.__is_serializable__(value):
            return value, type(value).__name__
        elif isinstance(value, Serializable):
            return value.serialize(), type(value).__name__
        try:
            pickler = jsonpickle.pickler.Pickler(max_depth=SERIALIZATION_DEPTH)
            value_dict = pickler.flatten(value, reset=True)
            return value_dict, type(value).__name__
        except:
            return '<not-json-serializable-object>', type(value).__name__

    def __call__(self, original_func):
        @wraps(original_func)
        def ___thundra_trace___(*args, **kwargs):
            parent_scope = self.tracer.scope_manager.active
            parent_span = parent_scope.span if parent_scope is not None else None

            # Set finish_on_close to False, otherwise it calls span's finish method which can raise an error,
            # which would cause the scope left open  
            scope = self.tracer.start_active_span(original_func.__name__, child_of=parent_span, finish_on_close=False)
            scope.span.class_name = 'Method'
            traced_err = None
            global _line_traced_count
            global DATA_LIMIT
            
            try:
                # Add argument related tags to the span before calling original method
                if self._trace_args is True:
                    function_args_list = []
                    count = 0
                    for arg in args:
                        value, value_type = self.__serialize_value__(arg)
                        function_args_dict = {
                            'name': 'arg-' + str(count),
                            'type': value_type,
                            'value': value
                        }
                        count += 1
                        function_args_list.append(function_args_dict)
                    if kwargs is not None:
                        for key, value in kwargs.items():
                            value, value_type = self.__serialize_value__(value)
                            function_args_dict = {
                                'name': key,
                                'type': value_type,
                                'value': value
                            }
                            function_args_list.append(function_args_dict)
                    scope.span.set_tag(constants.LineByLineTracingTags['args'], function_args_list)
                # Inform that span is initalized
                scope.span.on_started()
                self._tracing = True
                # Check if line-by-line tracing enabled
                if self.trace_line_by_line:
                    try:
                        if self.trace_lines_with_source:
                            source_lines, start_line = inspect.getsourcelines(original_func)
                            scope.span.set_tag(constants.LineByLineTracingTags['source'], ''.join(source_lines))
                            scope.span.set_tag(constants.LineByLineTracingTags['start_line'], start_line)
                    except Exception as e:
                        debug_logger("Cannot get source code in traceable: " + str(e))
                    with _lock:
                        if _line_traced_count == 0:
                            sys.settrace(trace_calls)
                        _line_traced_count += 1

                trace_local.trace_call_active = True
                # Call original func
                response = original_func(*args, **kwargs)
                self._tracing = False
                # Add return value related tags after having called the original func
                if self._trace_return_value is True and response is not None:
                    value, value_type = self.__serialize_value__(response)
                    return_value = {
                        'type': value_type,
                        'value': value
                    }
                    scope.span.set_tag('method.return_value', return_value)
            except Exception as e:
                if self._trace_error is True:
                    traced_err = e
                raise e
            finally:
                self._tracing = False
                # Disable line-by-line tracing if it is not used
                if self._trace_line_by_line:
                    with _lock:
                        _line_traced_count -= 1
                        if _line_traced_count == 0:
                            sys.settrace(None)

                try:
                    # Since span's finish method calls listeners, it can raise an error
                    scope.span.finish()
                except Exception as injected_err:
                    if traced_err is None:
                        traced_err = injected_err
                # Close the scope regardless of an error is raised or not
                scope.close()
                # Set the error tags to the span (if any)
                if traced_err is not None:
                    scope.span.set_error_to_tag(traced_err)
                    raise traced_err
            return response

        return ___thundra_trace___

    call = __call__



from thundra.wrappers.tornado.tornado_wrapper import TornadoWrapper, logger


class ThundraMiddleware(object):
    def __init__(self):
        super(ThundraMiddleware, self).__init__()
        self._wrapper = TornadoWrapper()

    def execute(self, handler): 
        try:
            request = handler.request
            self._wrapper.before_request(request)
        except Exception as e:
            logger.error("Error during the before part of Thundra: {}".format(e))

    def finish(self, handler, error=None): 
        try:
            http_response = type('', (object,), {"status_code": handler.get_status()})()
            self._wrapper.after_request(http_response, error)
        except Exception as e:
            logger.error("Error during the after part of Thundra: {}".format(e))



import abc

ABC = abc.ABCMeta('ABC', (object,), {})


class SpanFilterer(ABC):
    @abc.abstractmethod
    def accept(self, span):
        """
        Returns whether or not given span is able to pass through
        the filterer.
        """


class SpanFilter(ABC):
    @abc.abstractmethod
    def accept(self, span):
        """
        Returns whether or not given span is able to pass through
        the filter.
        """


class StandardSpanFilterer(SpanFilterer):

    def __init__(self, span_filters=None, all_mandatory=False):
        if span_filters is None:
            span_filters = []
        self.span_filters = span_filters
        self.all_mandatory = all_mandatory

    def accept(self, span):
        if (not self.span_filters) or len(self.span_filters) == 0:
            return True

        if self.all_mandatory:
            for sf in self.span_filters:
                if hasattr(sf, 'accept'):
                    if not sf.accept(span):
                        return False
                else:
                    raise TypeError("{} doesn't have accept method".format(type(sf)))
            return True

        else:
            for sf in self.span_filters:
                if hasattr(sf, 'accept'):
                    if sf.accept(span):
                        return True
                else:
                    raise TypeError("{} doesn't have accept method".format(type(sf)))
            return False

    def add_filter(self, span_filter):
        self.span_filters.append(span_filter)

    def clear_filters(self):
        self.span_filters = []


class SimpleSpanFilter(SpanFilter):

    def __init__(self,
                 domain_name=None,
                 class_name=None,
                 operation_name=None,
                 tags=None, reverse=False):
        self.domain_name = domain_name
        self.class_name = class_name
        self.operation_name = operation_name
        self.tags = tags
        self.reverse = reverse

    def accept(self, span):
        accepted = True
        if self.domain_name is not None:
            accepted = self.domain_name == span.domain_name

        if accepted and self.class_name is not None:
            accepted = self.class_name == span.class_name

        if accepted and self.operation_name is not None:
            accepted = self.operation_name == span.operation_name

        if accepted and self.tags is not None:
            for k, v in self.tags.items():
                if isinstance(v, list):
                    if not (span.get_tag(k) in v):
                        accepted = False
                        break
                elif span.get_tag(k) != v:
                    accepted = False
                    break
        if self.reverse:
            accepted = not accepted

        return accepted

    def __repr__(self):
        return ("SpanFilter with class_name={}, domain_name={}," +
                "operation_name={}, tags={}, reverse={}").format(self.class_name, self.domain_name, self.operation_name,
                                                                 self.tags, self.reverse)

    @staticmethod
    def from_config(config):
        kwargs = {}
        tags = config.get('tags')
        domain_name = config.get('domainName')
        class_name = config.get('className')
        operation_name = config.get('operationName')
        reverse = config.get('reverse')

        if domain_name is not None:
            kwargs['domain_name'] = str(domain_name)
        if class_name is not None:
            kwargs['class_name'] = str(class_name)
        if operation_name is not None:
            kwargs['operation_name'] = str(operation_name)
        if reverse is not None:
            kwargs['reverse'] = reverse
        if len(tags) > 0:
            kwargs['tags'] = tags

        return SimpleSpanFilter(**kwargs)



THUNDRA_MARKED_AS_SKIPPED = "thundra_marked_as_skipped"
THUNDRA_TEST_ALREADY_FINISHED = "thundra_test_already_finished"
THUNDRA_TEST_STARTED = "thundra_test_started"
THUNDRA_TEST_RESULTED = "thundra_test_resulted"
THUNDRA_TEST_FINISH_IN_HELPER = "thundra_test_finish_in_helper"
THUNDRA_FIXTURE_PREFIX = "x_thundra"
THUNDRA_SCOPE = "x-thundra-scope"


from foresight.environment.git.git_helper import GitHelper
from foresight.environment.environment_info import EnvironmentInfo
from foresight.utils.test_runner_utils import TestRunnerUtils
import os, logging
from foresight.utils.generic_utils import print_debug_message_to_console

LOGGER = logging.getLogger(__name__)

class CircleCIEnvironmentInfoProvider:
    ENVIRONMENT = "CircleCI"
    CIRCLE_REPOSITORY_URL_ENV_VAR_NAME = "CIRCLE_REPOSITORY_URL"
    CIRCLE_BRANCH_ENV_VAR_NAME = "CIRCLE_BRANCH"
    CIRCLE_SHA1_ENV_VAR_NAME = "CIRCLE_SHA1"
    CIRCLE_BUILD_URL_ENV_VAR_NAME = "CIRCLE_BUILD_URL"
    CIRCLE_BUILD_NUM_ENV_VAR_NAME = "CIRCLE_BUILD_NUM"

    @classmethod
    def get_test_run_id(cls, repo_url, commit_hash):
        configured_test_run_id = TestRunnerUtils.get_configured_test_run_id()
        if configured_test_run_id:
            return configured_test_run_id
        build_url = os.getenv(cls.CIRCLE_BUILD_URL_ENV_VAR_NAME)
        build_num = os.getenv(cls.CIRCLE_BUILD_NUM_ENV_VAR_NAME)
        test_run_key = TestRunnerUtils.string_concat_by_underscore(build_url, build_num)
        if test_run_key != "":
            return TestRunnerUtils.get_test_run_id(cls.ENVIRONMENT, repo_url, commit_hash,
                test_run_key)
        else:
            return TestRunnerUtils.get_default_test_run_id(cls.ENVIRONMENT, repo_url, commit_hash)
            

    @classmethod
    def build_env_info(cls):
        try:
            repo_url = os.getenv(cls.CIRCLE_REPOSITORY_URL_ENV_VAR_NAME)
            repo_name = GitHelper.extractRepoName(repo_url)
            branch = os.getenv(cls.CIRCLE_BRANCH_ENV_VAR_NAME)
            commit_hash = os.getenv(cls.CIRCLE_SHA1_ENV_VAR_NAME)
            commit_message = GitHelper.get_commit_message()

            if not branch:
                branch = GitHelper.get_branch()

            if not commit_hash:
                commit_hash = GitHelper.get_commit_hash()

            test_run_id = cls.get_test_run_id(repo_url, commit_hash)

            env_info = EnvironmentInfo(test_run_id, cls.ENVIRONMENT, repo_url, repo_name, branch, 
                commit_hash, commit_message)
            print_debug_message_to_console("CircleCI Environment info: {}".format(env_info.to_json()))
            return env_info
        except Exception as err:
            print_debug_message_to_console("CircleCI Unable to build environment info: {}".format(err))
            LOGGER.error("CircleCI Unable to build environment info: {}".format(err))
            pass
        return None


from thundra.samplers.base_sampler import BaseSampler
import fastcounter

class MaxCountAwareSampler(BaseSampler):

    def __init__(self, max_count):
        self.max_count = max_count
        self.counter = fastcounter.FastWriteCounter()

    def is_sampled(self, args=None):
        self.counter.increment()
        return self.counter.value <= self.max_count


from thundra.plugins.config.base_plugin_config import BasePluginConfig


class MetricConfig(BasePluginConfig):
    def __init__(self, opts=None):
        if opts is None:
            opts = {}
        super(MetricConfig, self).__init__(enabled=opts.get('enabled', True))
        self.sampler = opts.get('sampler')



import logging
from foresight.environment.git.git_helper import GitHelper
from foresight.environment.environment_info import EnvironmentInfo
from foresight.utils.test_runner_utils import TestRunnerUtils
from foresight.utils.generic_utils import print_debug_message_to_console

LOGGER = logging.getLogger(__name__)

class GitEnvironmentInfoProvider:
    
    ENVIRONMENT = "Git"

    @classmethod
    def get_test_run_id(cls, repo_url, commit_hash):
        configured_test_run_id = TestRunnerUtils.get_configured_test_run_id()
        if configured_test_run_id:
            return configured_test_run_id
        return TestRunnerUtils.get_default_test_run_id(cls.ENVIRONMENT, repo_url, commit_hash)

    @classmethod
    def build_env_info(cls):
        try:
            repo_url = GitHelper.get_repo_url()
            if not repo_url:
                return None
            repo_name = GitHelper.extractRepoName(repo_url)
            branch = GitHelper.get_branch()
            commit_hash = GitHelper.get_commit_hash()
            commit_message = GitHelper.get_commit_message()
            test_run_id = cls.get_test_run_id(repo_url, commit_hash)
            env_info = EnvironmentInfo(test_run_id, cls.ENVIRONMENT, repo_url, repo_name,
                branch, commit_hash, commit_message)
            print_debug_message_to_console("Git Environment info: {}".format(env_info.to_json()))
            return env_info
        except Exception as err:
            print_debug_message_to_console("Unable to build environment info: {}".format(err))
            LOGGER.error("Unable to build environment info: {}".format(err))
            pass
        return None


from thundra import constants
from thundra.wrappers import wrapper_utils, web_wrapper_utils

from thundra.wrappers.fastapi.fastapi_utils import extract_headers

import json

def start_trace(plugin_context, execution_context, tracer):
    request = execution_context.platform_data["request"]

    request = {
        "method": request.get("method"),
        "host": request.get("server")[0],
        "query_params": request.get("query_string"),
        "headers": extract_headers(request),
        "body": request.get("body"),
        "path": request.get("path")
    }
    
    web_wrapper_utils.start_trace(execution_context, tracer, "Fastapi", "API", request)


def finish_trace(execution_context):
    root_span = execution_context.root_span
    if execution_context.response:
        status_code = get_response_status(execution_context)
        if status_code:
            root_span.set_tag(constants.HttpTags['HTTP_STATUS'], status_code)
        if execution_context.trigger_operation_name:
            if isinstance(execution_context.response, dict):
                if execution_context.response.get('headers'):
                    execution_context.response.get('headers')[
                        constants.TRIGGER_RESOURCE_NAME_KEY] = execution_context.trigger_operation_name
            else:
                if hasattr(execution_context.response, 'headers'):
                    execution_context.response.headers[
                        constants.TRIGGER_RESOURCE_NAME_KEY] = execution_context.trigger_operation_name
    web_wrapper_utils.finish_trace(execution_context)
    

def start_invocation(plugin_context, execution_context):
    execution_context.invocation_data = wrapper_utils.create_invocation_data(plugin_context, execution_context)


def finish_invocation(execution_context):
    wrapper_utils.finish_invocation(execution_context)

    wrapper_utils.set_response_status(execution_context, get_response_status(execution_context))


def get_response_status(execution_context):
    try:
        status_code = execution_context.response.get("status_code") or execution_context.response.get("status")
    except:
        return None
    return status_code


from functools import wraps
from thundra.opentracing.tracer import ThundraTracer


class TraceAwareWrapper:

    def __init__(self, active_span=None):
        if active_span is None:
            active_scope = ThundraTracer.get_instance().scope_manager.active
            active_span = active_scope.span if active_scope is not None else None
        self._parent_span = active_span

    @property
    def parent_span(self):
        return self._parent_span

    def __call__(self, original_func):
        @wraps(original_func)
        def wrapper(*args, **kwargs):
            ThundraTracer.get_instance().scope_manager.activate(self.parent_span, True)
            return original_func(*args, **kwargs)
        return wrapper



from thundra.application.application_info_provider import ApplicationInfoProvider
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider


class GlobalApplicationInfoProvider(ApplicationInfoProvider):
    def __init__(self, application_info_provider=None):
        self.application_info = {}
        self.application_info_provider = application_info_provider
        if self.application_info_provider:
            self.application_info = self.application_info_provider.get_application_info()

        app_info_from_config = self.get_application_info_from_config()

        self.update(app_info_from_config)

    def get_application_info(self):
        return self.application_info

    def get_application_tags(self):
        return self.application_info.get('applicationTags', {}).copy()

    @staticmethod
    def get_application_info_from_config():
        return {
            'applicationId': ConfigProvider.get(config_names.THUNDRA_APPLICATION_ID),
            'applicationInstanceId': ConfigProvider.get(config_names.THUNDRA_APPLICATION_INSTANCE_ID),
            'applicationDomainName': ConfigProvider.get(config_names.THUNDRA_APPLICATION_DOMAIN_NAME),
            'applicationClassName': ConfigProvider.get(config_names.THUNDRA_APPLICATION_CLASS_NAME),
            'applicationName': ConfigProvider.get(config_names.THUNDRA_APPLICATION_NAME),
            'applicationVersion': ConfigProvider.get(config_names.THUNDRA_APPLICATION_VERSION, ''),
            'applicationStage': ConfigProvider.get(config_names.THUNDRA_APPLICATION_STAGE, ''),
            'applicationRegion': ConfigProvider.get(config_names.THUNDRA_APPLICATION_REGION, ''),
            'applicationRuntime': ApplicationInfoProvider.APPLICATION_RUNTIME,
            'applicationRuntimeVersion': ApplicationInfoProvider.APPLICATION_RUNTIME_VERSION,
            'applicationTags': ApplicationInfoProvider.parse_application_tags()
        }

    def update(self, opts):
        filtered_opts = {k: v for k, v in opts.items() if v is not None}
        self.application_info.update(filtered_opts)



from thundra.listeners.thundra_span_filterer import SpanFilter


class CompositeSpanFilter(SpanFilter):
    
    def __init__(self, is_all=False, filters=None):
        if filters is None:
            filters = []
        self.all = is_all
        self.filters = filters

    def accept(self, span):
        if (not self.filters) or len(self.filters) == 0:
            return True

        result = self.all

        if self.all:
            for span_filter in self.filters:
                result = result and span_filter.accept(span)
        else:
            for span_filter in self.filters:
                result = result or span_filter.accept(span)
        return result
    
    def set_filters(self, filters):
        self.filters = filters



import __future__
import os
import logging
from foresight.environment.git.utils import backward_search_for_file

LOGGER = logging.getLogger(__name__)

class GitHelper:

    USER_DIR = os.getcwd()
    SOURCE_CODE_PATH = "sourceCodePath"
    REPOSITORY_URL = "repositoryURL"
    COMMIT_HASH = "commitHash"
    COMMIT_MESSAGE = "commitMessage"
    BRANCH = "branch"
    GIT_FOLDER_NAME = ".git"
    git_info_map = dict()


    @classmethod
    def get_source_root_path(cls):
        return cls.git_info_map.get(GitHelper.SOURCE_CODE_PATH, None)


    @classmethod
    def get_repo_url(cls):
        return cls.git_info_map.get(GitHelper.REPOSITORY_URL, None)


    @classmethod
    def get_branch(cls):
        return cls.git_info_map.get(GitHelper.BRANCH, None)


    @classmethod
    def get_commit_hash(cls):
        return cls.git_info_map.get(GitHelper.COMMIT_HASH, None)


    @classmethod
    def get_commit_message(cls):
        return cls.git_info_map.get(GitHelper.COMMIT_MESSAGE, None)


    @classmethod
    def populate_git_info_map(cls):
        git_folder_path = backward_search_for_file(cls.USER_DIR, cls.GIT_FOLDER_NAME)
        if not git_folder_path:
            LOGGER.debug("Could not locate " + cls.USER_DIR + " starting from user.dir: " + cls.GIT_FOLDER_NAME)
            cls.git_info_map = {}
            return
        try:
            from git import Repo
            git_repo = Repo(git_folder_path)
            active_branch = git_repo.active_branch
            commit_hash = str(active_branch.commit)
            commit_message = active_branch.commit.message
            repo_url = git_repo.remotes[0].config_reader.get("url")
            source_code_path = git_repo.working_dir
            cls.git_info_map[GitHelper.SOURCE_CODE_PATH] = source_code_path
            cls.git_info_map[GitHelper.BRANCH] = active_branch.name
            cls.git_info_map[GitHelper.COMMIT_HASH] = commit_hash
            cls.git_info_map[GitHelper.COMMIT_MESSAGE] = commit_message
            cls.git_info_map[GitHelper.REPOSITORY_URL] = repo_url
        except Exception as err:
            LOGGER.error("Couldn't set git_info_map: {}: ".format(err))
            pass

    @staticmethod
    def extractRepoName(repo_url):
        try:
            return os.path.splitext(os.path.basename(repo_url))[0]
        except Exception as err:
            LOGGER.error("Couldn't extract Repo Name: {}".format(err))
            pass

GitHelper.populate_git_info_map()


from thundra.context.execution_context import ExecutionContext
from thundra.context.global_execution_context_provider import GlobalExecutionContextProvider


class ExecutionContextManager:
    context_provider = None

    @staticmethod
    def set_provider(provider):
        ExecutionContextManager.context_provider = provider

    @staticmethod
    def set(context):
        ExecutionContextManager.context_provider.set(context)

    @staticmethod
    def clear():
        ExecutionContextManager.context_provider.clear()

    @staticmethod
    def get():
        if not ExecutionContextManager.context_provider:
            return ExecutionContext()
        execution_context = ExecutionContextManager.context_provider.get()
        if not execution_context:
            return ExecutionContext()
        return execution_context



import wrapt


class CursorWrapper(wrapt.ObjectProxy):

    def __init__(self, cursor, connection_wrapper, integration):
        super(CursorWrapper, self).__init__(cursor)
        self._self_connection = connection_wrapper
        self.integration = integration

    def execute(self, *args, **kwargs):
        return self.integration.run_and_trace(
            self.__wrapped__.execute,
            self._self_connection,
            args,
            kwargs,
        )

    def callproc(self, *args, **kwargs):
        return self.integration.run_and_trace(
            self.__wrapped__.callproc,
            self._self_connection,
            args,
            kwargs,
        )

    def __enter__(self):
        # raise appropriate error if api not supported (should reach the user)
        self.__wrapped__.__enter__
        return self


class ConnectionWrapper(wrapt.ObjectProxy):
    def __init__(self, conn, integration=None):
        self.integration = integration

    def cursor(self, *args, **kwargs):
        cursor = self.__wrapped__.cursor(*args, **kwargs)
        return CursorWrapper(cursor, self, integration=self.integration)



from thundra.context.execution_context_manager import ExecutionContextManager


def set_agent_tag(key, value):
    execution_context = ExecutionContextManager.get()
    execution_context.tags[key] = value


def set_many_agent(tags):
    execution_context = ExecutionContextManager.get()
    execution_context.tags.update(tags)


def get_agent_tag(key):
    execution_context = ExecutionContextManager.get()
    if key in execution_context.tags:
        return execution_context.tags[key]
    return None


def get_agent_tags():
    execution_context = ExecutionContextManager.get()
    return execution_context.tags.copy()


def remove_agent_tag(key):
    execution_context = ExecutionContextManager.get()
    return execution_context.tags.pop(key, None)


def set_tag(key, value):
    execution_context = ExecutionContextManager.get()
    execution_context.user_tags[key] = value


def set_tags(tags):
    execution_context = ExecutionContextManager.get()
    execution_context.user_tags.update(tags)


def set_many(tags):
    execution_context = ExecutionContextManager.get()
    execution_context.user_tags.update(tags)


def get_tag(key):
    execution_context = ExecutionContextManager.get()
    if key in execution_context.user_tags:
        return execution_context.user_tags[key]
    return None


def get_tags():
    execution_context = ExecutionContextManager.get()
    return execution_context.user_tags.copy()


def remove_tag(key):
    execution_context = ExecutionContextManager.get()
    return execution_context.user_tags.pop(key, None)


def clear():
    execution_context = ExecutionContextManager.get()
    execution_context.user_tags.clear()
    execution_context.tags.clear()


def clear_error():
    execution_context = ExecutionContextManager.get()
    execution_context.user_error = None


def set_error(err):
    execution_context = ExecutionContextManager.get()
    execution_context.user_error = err


def get_error():
    execution_context = ExecutionContextManager.get()
    return execution_context.user_error


def set_application_resource_name(resource_name):
    execution_context = ExecutionContextManager.get()
    execution_context.application_resource_name = resource_name



import os
from thundra import utils, constants


def test_get_default_timeout_margin(monkeypatch):
    monkeypatch.setitem(os.environ, constants.AWS_REGION, 'us-west-2')
    timeout_margin = utils.get_default_timeout_margin()
    assert timeout_margin == 200

    monkeypatch.setitem(os.environ, constants.AWS_REGION, 'us-east-2')

    timeout_margin = utils.get_default_timeout_margin()
    assert timeout_margin == 600

    monkeypatch.setitem(os.environ, constants.AWS_REGION, 'eu-west-1')

    timeout_margin = utils.get_default_timeout_margin()
    assert timeout_margin == 1000

    monkeypatch.setitem(os.environ, constants.AWS_REGION, 'eu-west-1')
    monkeypatch.setitem(os.environ, constants.AWS_LAMBDA_FUNCTION_MEMORY_SIZE, '1536')

    timeout_margin = utils.get_default_timeout_margin()
    assert timeout_margin == 1000

    monkeypatch.setitem(os.environ, constants.AWS_REGION, 'eu-west-1')
    monkeypatch.setitem(os.environ, constants.AWS_LAMBDA_FUNCTION_MEMORY_SIZE, '128')

    timeout_margin = utils.get_default_timeout_margin()
    assert timeout_margin == 3000


def test_get_nearest_collector(monkeypatch):
    regions = [ 'us-west-2', 'us-west-1', 'us-east-2', 'us-east-1', 'ca-central-1', 'sa-east-1',
            'eu-central-1', 'eu-west-1', 'eu-west-2', 'eu-west-3', 'eu-north-1', 'eu-south-1',
            'ap-south-1', 'ap-northeast-1', 'ap-northeast-2', 'ap-southeast-1', 'ap-southeast-2', 
            'ap-east-1', 'af-south-1', 'me-south-1']

    for region in regions:
        monkeypatch.setitem(os.environ, constants.AWS_REGION, region)
        collector = utils.get_nearest_collector()
        assert collector == "{}.collector.thundra.io".format(region)
    
    monkeypatch.delitem(os.environ, constants.AWS_REGION)
    collector = utils.get_nearest_collector()
    assert collector == "collector.thundra.io"



import time
import uuid

from thundra import constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.context.execution_context import ExecutionContext
from thundra.plugins.invocation import invocation_trace_support
from thundra.plugins.invocation.invocation_plugin import InvocationPlugin
from thundra.plugins.log.log_plugin import LogPlugin
from thundra.plugins.metric.metric_plugin import MetricPlugin
from thundra.plugins.trace.trace_plugin import TracePlugin


def initialize_plugins(plugin_context, disable_trace, disable_metric, disable_log, config):
    plugins = []
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_DISABLE, disable_trace):
        plugins.append(TracePlugin(plugin_context=plugin_context, config=config.trace_config))
    plugins.append(InvocationPlugin(plugin_context=plugin_context))

    if not ConfigProvider.get(config_names.THUNDRA_METRIC_DISABLE, disable_metric):
        plugins.append(MetricPlugin(plugin_context=plugin_context, config=config.metric_config))

    if not ConfigProvider.get(config_names.THUNDRA_LOG_DISABLE, disable_log):
        plugins.append(LogPlugin(plugin_context=plugin_context, config=config.log_config))
    return plugins


def create_invocation_data(plugin_context, execution_context):
    if not execution_context.transaction_id:
        execution_context.transaction_id = str(uuid.uuid4())
    invocation_data = {
        'id': str(uuid.uuid4()),
        'type': "Invocation",
        'agentVersion': constants.THUNDRA_AGENT_VERSION,
        'dataModelVersion': constants.DATA_FORMAT_VERSION,
        'traceId': execution_context.trace_id,
        'transactionId': execution_context.transaction_id,
        'spanId': execution_context.span_id,
        'applicationPlatform': '',
        'applicationRegion': plugin_context.application_info.get('applicationRegion'),
        'duration': None,
        'startTimestamp': execution_context.start_timestamp,
        'finishTimestamp': None,
        'erroneous': False,
        'errorType': '',
        'errorMessage': '',
        'errorStack': '',
        'errorCode': -1,
        'coldStart': plugin_context.request_count == 1,
        'timeout': False,
        'tags': {},
    }

    # Add application related data
    application_info = plugin_context.application_info
    invocation_data.update(application_info)

    return invocation_data


def finish_invocation(execution_context):
    invocation_data = execution_context.invocation_data

    # Add user tags
    invocation_data['userTags'] = execution_context.user_tags

    # Add agent tags
    invocation_data['tags'].update(execution_context.tags)

    # Get resources
    resources = invocation_trace_support.get_resources()
    invocation_data.update(resources)

    # Get incoming trace links
    incoming_trace_links = invocation_trace_support.get_incoming_trace_links()
    invocation_data.update(incoming_trace_links)

    # Get outgoing trace links
    outgoing_trace_links = invocation_trace_support.get_outgoing_trace_links()
    invocation_data.update(outgoing_trace_links)

    # Check errors
    user_error = execution_context.user_error
    if execution_context.error:
        set_error(invocation_data, execution_context.error)
    elif user_error:
        set_error(invocation_data, user_error)

    duration = execution_context.finish_timestamp - execution_context.start_timestamp
    invocation_data['duration'] = int(duration)
    invocation_data['finishTimestamp'] = int(execution_context.finish_timestamp)
    invocation_data['timeout'] = execution_context.timeout
    invocation_data['applicationResourceName'] = execution_context.application_resource_name

    execution_context.invocation_data = invocation_data


def set_error(invocation_data, error):
    invocation_data['erroneous'] = True
    if isinstance(error, Exception):
        error_type = type(error)
        invocation_data['errorType'] = error_type.__name__
        invocation_data['errorMessage'] = str(error)
        if hasattr(error, 'code'):
            invocation_data['errorCode'] = error.code
    elif isinstance(error, dict):
        invocation_data['errorType'] = error.get('type')
        invocation_data['errorMessage'] = error.get('message')
        invocation_data['errorStack'] = error.get('traceback', None)


def create_execution_context():
    transaction_id = str(uuid.uuid4())
    start_timestamp = int(time.time() * 1000)
    return ExecutionContext(transaction_id=transaction_id, start_timestamp=start_timestamp)


def set_response_status(execution_context, response_status_code):
    if response_status_code:
        if execution_context.invocation_data.get('userTags') is None:
            execution_context.invocation_data['userTags'] = {}
        if execution_context.invocation_data['userTags'].get('http.status_code') is None:
            execution_context.invocation_data['userTags']['http.status_code'] = response_status_code



from thundra import constants
from thundra.wrappers import wrapper_utils, web_wrapper_utils


def start_trace(plugin_context, execution_context, tracer):
    request = execution_context.platform_data['request']
    import sys
    if sys.version_info[0] >= 3:
        request_route_path = str(
            request.resolver_match.route) if request.resolver_match and request.resolver_match.route else None
        request = {
            'method': request.method,
            'host': request.get_host().split(':')[0],
            'query_params': request.GET,
            'body': request.body,
            'headers': request.headers,
            'path': request.path
        }
    else:
        request_route_path = str(
            request.resolver_match.url_name) if request.resolver_match and request.resolver_match.url_name else None
        request = {
            'method': request.method,
            'host': request.get_host().split(':')[0],
            'query_params': request.GET,
            'body': request.body,
            'headers': request.META,
            'path': request.path
        }
    web_wrapper_utils.start_trace(execution_context, tracer, 'Django', 'API', request, request_route_path)


def finish_trace(execution_context):
    root_span = execution_context.root_span
    if execution_context.response:
        root_span.set_tag(constants.HttpTags['HTTP_STATUS'], execution_context.response.status_code)
        if execution_context.trigger_operation_name:
            execution_context.response[
                constants.TRIGGER_RESOURCE_NAME_KEY] = execution_context.trigger_operation_name
    web_wrapper_utils.finish_trace(execution_context)


def start_invocation(plugin_context, execution_context):
    execution_context.invocation_data = wrapper_utils.create_invocation_data(plugin_context, execution_context)


def finish_invocation(execution_context):
    wrapper_utils.finish_invocation(execution_context)

    # Set response status code
    wrapper_utils.set_response_status(execution_context, get_response_status(execution_context))


def get_response_status(execution_context):
    try:
        status_code = execution_context.response.status_code
    except:
        return None
    return status_code



import pytest, logging, os

from foresight.pytest_integration.utils import (patch, check_test_case_result, 
    update_test_status, set_attributes_test_item, check_test_status_state)
from foresight.pytest_integration.pytest_helper import PytestHelper
from foresight import foresight_executor
import foresight.pytest_integration.constants as pytest_constants
from thundra.context.execution_context_manager import ExecutionContextManager

logger = logging.getLogger(__name__)


def pytest_addoption(parser):
    """Setup thundra for pytest. Adding both ini file and command line argument. 

    Args:
        parser (pytest.Parser): Parser for command line arguments and ini-file values
    """
    try:
        group = parser.getgroup("thundra")

        # config.getoption(markerName) 
        group.addoption(
            "--thundra_disable",
            action="store_true",
            dest="thundra_disable",
            default=False,
            help='THUNDRA_APIKEY and THUNDRA_AGENT_TEST_PROJECT_ID should be set as environment variables. \
                Disable thundra by --thundra_disable argument. ',
        )
        parser.addini("thundra_disable", 'Disable tracing of pytest functions by --thundra_disable argument', type="bool")
    except Exception as e:
        logger.error("Pytest addoption error: {}".format(e))


def check_thundra_from_env():
    check_thundra_api_key = None
    check_thundra_project_id = None
    for key in os.environ.keys():
        lower_key = key.lower()
        if lower_key.startswith("thundra_"):
            if lower_key == "thundra_apikey":
                check_thundra_api_key = True
            elif lower_key == "thundra_agent_test_project_id":
                check_thundra_project_id = True
            if check_thundra_api_key and check_thundra_project_id:
                break
    return check_thundra_api_key, check_thundra_project_id

def check_thundra_test_disabled():
    check_thundra_test_disabled = False
    for key in os.environ.keys():
        lower_key = key.lower()
        if lower_key.startswith("thundra_"):
            if lower_key == "thundra_agent_test_disable":
                check_thundra_test_disabled = False if os.getenv(key).lower() == "false" else True
                break
    return check_thundra_test_disabled

'''
    if thundra is imported in conftest!!!
    example:
        import thundra
        thundra.configure(
        options={
            "config": {
                "thundra.apikey": <test_id>,
                "thundra.agent.test.project.id": <project_id>,
            }
        }
)
'''
def check_thundra_from_conf():
    from thundra.config.config_provider import ConfigProvider
    from thundra.config import config_names
    api_key = ConfigProvider.get(config_names.THUNDRA_APIKEY, None)
    project_id = ConfigProvider.get(config_names.THUNDRA_TEST_PROJECT_ID, None)
    return api_key, project_id


def pytest_load_initial_conftests(early_config, parser, args):
    """Check --collect-only args set by user. If it's set, then there will no any test run. 
    Therefore, no need to start Thundra.

    Args:
        early_config (Config): Pytest Config Object
        parser (Parser): Pytest Parser Object
        args (List[str]): Pytest command line args
    """
    if "--collect-only" in args:
        PytestHelper.set_pytest_collect_only()

def pytest_sessionstart(session):
    """ Check thundra has been activated. If it has been, then start session.

    Args:
        session (pytest.Session): Pytest session class
    """
    try:
        from thundra.config.config_provider import ConfigProvider
        from thundra.config import config_names
        if (session.config.getoption("thundra_disable") or 
            session.config.getini("thundra_disable") or 
            check_thundra_test_disabled() or 
            ConfigProvider.get(config_names.THUNDRA_TEST_DISABLE, False)):
            import thundra 
            already_configured = True if ConfigProvider.configs else False
            thundra._set_thundra_for_test_env(already_configured)
        else:
            api_key_env, project_id_env = check_thundra_from_env()
            api_key_conf, project_id_conf = check_thundra_from_conf()
            check_thundra_project_id = project_id_env or project_id_conf
            check_thundra_api_key = api_key_env or api_key_conf
            if check_thundra_project_id and check_thundra_api_key and not PytestHelper.check_pytest_collect_only():
                PytestHelper.set_pytest_started()
                patch()
                PytestHelper.session_setup(executor=foresight_executor)
            else:
                print("[THUNDRA] Please make sure setting THUNDRA_APIKEY and THUNDRA_AGENT_TEST_PROJECT_ID as environment variables and pytest --collect-only args ain't activated!")
    except Exception as e:
        logger.error("Pytest pytest_sessionstart error: {}".format(e))
        pass

def pytest_sessionfinish(session, exitstatus):
    """ Finish session for thundra

    Args:
        session (pytest.Session): Pytest session class
        exitstatus (int): The status which pytest will return to the system
    """
    try:
        if PytestHelper.check_pytest_started():
            PytestHelper.session_teardown()
    except Exception as e:
        logger.error("Pytest sessionfinish error: {}".format(e))
        pass

@pytest.fixture(scope="function", autouse=True)
def x_thundra_finish_test(request):
    """ If test has a fixture, then finish test case after test case body and 
    all fixture_teardown has been done.

    Args:
        request ([type]): [description]
    """
    try:
        if PytestHelper.check_pytest_started():
            yield
            PytestHelper.finish_test_span(request.node)
        else:
            yield
    except Exception as e:
        logger.error("Pytest x_thundra_finish_test error: {}".format(e))
        pass

# Perform the runtest protocol for a single test item
@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_protocol(item, nextitem):    
    """Perform the runtest protocol for a single test item. Set attributes for identify 
    item will be used for test suite or test case, and it has any skip marked. 
    Start test suite and after all flow completed and nextitem is None or current item.module not equal to
    nextitem.module finish test suite also.

    Args:
        item (pytest.Item): Test item for which the runtest protocol is performed.
        nextitem (pytest.Item | None):  The scheduled-to-be-next test item (or None if this is the end my friend).
    """
    try:
        if not PytestHelper.check_pytest_started():
            yield
            return
        module_item = item.getparent(pytest.Module)
        set_attributes_test_item(item, module_item)
        PytestHelper.start_test_suite_span(module_item)
        yield
        if not hasattr(item, pytest_constants.THUNDRA_TEST_FINISH_IN_HELPER):
            PytestHelper.finish_test_span(item)
        else:
            delattr(item, pytest_constants.THUNDRA_TEST_FINISH_IN_HELPER)
        if not nextitem or item.getparent(pytest.Module).nodeid != nextitem.getparent(pytest.Module).nodeid:
            PytestHelper.finish_test_suite_span()
    except Exception as e:
        logger.error("Pytest runtest_protocol error: {}".format(e))
        pass

# Called to create a _pytest.reports.TestReport for each of the setup, call and teardown runtest phases of a test item
@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """Called to create a _pytest.reports.TestReport for each of the setup, 
    call and teardown runtest phases of a test item. Check test case span started or not and
    its marked skipped or not. Then, update test status according to test result.

    Args:
        item (pytest.Item): Current running item(test case).
        call (pytest.CallInfo): The CallInfo for the phase.
    """
    try:
        outcome = yield
        if not PytestHelper.check_pytest_started():
            return
        if call.when == "setup":
                PytestHelper.start_test_span(item)
        status, exception = check_test_status_state(item, call)
        if not status:
            return
        result = outcome.get_result()
        execution_context = ExecutionContextManager.get()  
        # After Function call report to get test status(success, failed, aborted, skipped , ignored)
        test_status = check_test_case_result(item, execution_context, result, exception)
        update_test_status(item, test_status, execution_context)
    except Exception as e:
        logger.error("Pytest runtest_makereport error: {}".format(e))
        pass


from thundra.listeners.thundra_span_listener import ThundraSpanListener
from thundra.listeners.error_injector_span_listener import ErrorInjectorSpanListener
from thundra.listeners.latency_injector_span_listener import LatencyInjectorSpanListener
from thundra.listeners.filtering_span_listener import FilteringSpanListener
from thundra.listeners.tag_injector_span_listener import TagInjectorSpanListener
from thundra.listeners.security_aware_span_listener import SecurityAwareSpanListener



import json

from thundra.encoder import to_json
from thundra import constants


class InvocationPlugin:

    def __init__(self, plugin_context=None):
        self.hooks = {
            'before:invocation': self.before_invocation,
            'after:invocation': self.after_invocation
        }
        self.plugin_context = plugin_context

    def before_invocation(self, execution_context):
        executor = self.plugin_context.executor
        if executor:
            executor.start_invocation(self.plugin_context, execution_context)

    def after_invocation(self, execution_context):
        executor = self.plugin_context.executor
        if executor:
            executor.finish_invocation(execution_context)
        report_data = {
            'apiKey': self.plugin_context.api_key,
            'type': 'Invocation',
            'dataModelVersion': constants.DATA_FORMAT_VERSION,
            'data': execution_context.invocation_data
        }
        execution_context.report(json.loads(to_json(report_data)))



from foresight.model.test_run_result import TestRunResult
from foresight.model.test_run_monitoring import TestRunMonitoring

class TestRunFinish(TestRunResult, TestRunMonitoring):
    EVENT_NAME = "TestRunFinish"

    def __init__(self, id=None, project_id=None, task_id=None, total_count=None, successful_count=None, 
        failed_count=None, ignored_count=None, aborted_count=None, start_timestamp=None, finish_timestamp=None,
        duration=None, host_name=None, environment_info=None, tags=None):
        super(TestRunFinish, self).__init__(total_count, successful_count, failed_count,
            ignored_count, aborted_count)
        self.id = id
        self.project_id = project_id
        self.task_id = task_id
        self.start_timestamp = start_timestamp
        self.finish_timestamp = finish_timestamp
        self.duration = duration
        self.host_name = host_name
        if environment_info != None:
            self.environment = environment_info.environment if environment_info.environment else None
            self.repo_url = environment_info.repo_url if environment_info.repo_url else None
            self.repo_name = environment_info.repo_name if environment_info.repo_name else None
            self.branch = environment_info.branch if environment_info.branch else None
            self.commit_hash = environment_info.commit_hash if environment_info.commit_hash else None
            self.commit_message = environment_info.commit_message if environment_info.commit_message else None
        else:
            self.environment = None
            self.repo_url = None
            self.repo_name = None
            self.branch = None
            self.commit_hash = None
            self.commit_message = None
        self.tags = tags

    def to_json(self):
        return {
            "id": self.id,
            "projectId": self.project_id,
            "taskId": self.task_id,
            "type": self.EVENT_NAME,
            "agentVersion": self.AGENT_VERSION,
            "dataModelVersion": self.TEST_RUN_DATA_MODEL_VERSION,
            "totalCount" : self.total_count,
            "successfulCount" : self.successful_count,
            "failedCount" : self.failed_count,
            "ignoredCount" : self.ignored_count,
            "abortedCount" : self.aborted_count,
            "startTimestamp" : self.start_timestamp,
            "finishTimestamp" : self.finish_timestamp,
            "duration" : self.duration,
            "environment": self.environment,
            "repoURL": self.repo_url,
            "repoName": self.repo_name,
            "branch": self.branch,
            "commitHash": self.commit_hash,
            "commitMessage": self.commit_message,
            "tags": self.tags
        }

    def get_monitoring_data(self):
        monitoring_data = super().get_monitoring_data()
        monitoring_data["type"] = self.EVENT_NAME
        monitoring_data["data"] = self.to_json()
        return monitoring_data   


from opentracing import InvalidCarrierException

from thundra import constants
from thundra.opentracing.propagation.propagator import Propagator
from thundra.opentracing.span_context import ThundraSpanContext


class TextMapPropagator(Propagator):

    def inject(self, span_context, carrier):
        if span_context.trace_id:
            carrier[constants.THUNDRA_TRACE_ID_KEY] = span_context.trace_id

        if span_context.span_id:
            carrier[constants.THUNDRA_SPAN_ID_KEY] = span_context.span_id

        if span_context.transaction_id:
            carrier[constants.THUNDRA_TRANSACTION_ID_KEY] = span_context.transaction_id

        # Inject baggage items
        if span_context.baggage is not None:
            for key in span_context.baggage:
                carrier[constants.THUNDRA_BAGGAGE_PREFIX + key] = span_context.baggage[key]

    def extract(self, carrier):
        try:
            trace_id = carrier[constants.THUNDRA_TRACE_ID_KEY]
            span_id = carrier[constants.THUNDRA_SPAN_ID_KEY]
            transaction_id = carrier[constants.THUNDRA_TRANSACTION_ID_KEY]
        except:
            return None

        if not (trace_id and span_id and transaction_id):
            return None

        # Extract baggage items
        baggage = {}
        for key in carrier:
            if key.startswith(constants.THUNDRA_BAGGAGE_PREFIX):
                baggage[key[len(constants.THUNDRA_BAGGAGE_PREFIX):]] = carrier[key]

        span_context = ThundraSpanContext(trace_id=trace_id, span_id=span_id, transaction_id=transaction_id,
                                          baggage=baggage)

        return span_context



from foresight.test_runner_support import TestRunnerSupport
import logging

logger = logging.getLogger(__name__)

class TestStatus:
    SUCCESSFUL = "SUCCESSFUL"
    FAILED = "FAILED"
    ABORTED = "ABORTED"
    SKIPPED = "SKIPPED"
    TOTAL = "TOTAL"

def increase_successful_count():
    try:
        test_run_context = TestRunnerSupport.test_run_scope
        test_suite_context = TestRunnerSupport.test_suite_execution_context

        test_run_context.context.increase_successful_count()
        test_suite_context.increase_successful_count()
        _increase_total_count(test_run_context, test_suite_context)
    except Exception as err:
        logger.error("increase_successful_count error: {}".format(err))
        pass


def increase_failed_count():
    try:
        test_run_context = TestRunnerSupport.test_run_scope
        test_suite_context = TestRunnerSupport.test_suite_execution_context

        test_run_context.context.increase_failed_count()
        test_suite_context.increase_failed_count()
        _increase_total_count(test_run_context, test_suite_context)
    except Exception as err:
        logger.error("increase_failed_count error: {}".format(err))
        pass

def increase_aborted_count():
    try:
        test_run_context = TestRunnerSupport.test_run_scope
        test_suite_context = TestRunnerSupport.test_suite_execution_context

        test_run_context.context.increase_aborted_count()
        test_suite_context.increase_aborted_count()
        _increase_total_count(test_run_context, test_suite_context)
    except Exception as err:
        logger.error("increase_aborted_count error: {}".format(err))
        pass


def increase_skipped_count():
    try:
        test_run_context = TestRunnerSupport.test_run_scope
        test_suite_context = TestRunnerSupport.test_suite_execution_context

        test_run_context.context.increase_ignored_count()
        test_suite_context.increase_ignored_count() 
        _increase_total_count(test_run_context, test_suite_context)
    except Exception as err:
        logger.error("increase_skipped_count error: {}".format(err))
        pass


def _increase_total_count(test_run_context, test_suite_context):
    try:
        test_run_context.context.increase_total_count()
        test_suite_context.increase_total_count()
    except Exception as err:
        logger.error("_increase_total_count error: {}".format(err))
        pass

increase_actions = {
    TestStatus.SUCCESSFUL: increase_successful_count,
    TestStatus.FAILED: increase_failed_count,
    TestStatus.ABORTED: increase_aborted_count,
    TestStatus.SKIPPED: increase_skipped_count,
}


class TestRunResult:

    def __init__(self, total_count=0, successful_count=0, failed_count=0,
     ignored_count=0, aborted_count=0):
        self.total_count = total_count
        self.successful_count = successful_count
        self.failed_count = failed_count
        self.ignored_count = ignored_count
        self.aborted_count = aborted_count


    def to_json(self):
        return {
            "totalCount" : self.total_count,
            "successfulCount" : self.successful_count,
            "failedCount" : self.failed_count,
            "ignoredCount" : self.ignored_count,
            "abortedCount" : self.aborted_count,
        }


import os
import select
import socket
import sys

sys.path.append("/opt/python")
sys.path.append("/var/task")

import websocket

try:
    import thread
except ImportError:
    import _thread as thread

OPCODE_BINARY = 0x2
PROTOCOL_VER = 1.0
BROKER_HANDSHAKE_HEADERS = {
    "AUTH_TOKEN": "x-catchpoint-auth-token",
    "SESSION_NAME": "x-catchpoint-session-name",
    "PROTOCOL_VER": "x-catchpoint-protocol-version",
    "RUNTIME": "x-catchpoint-runtime",
    "SESSION_TIMEOUT": "x-catchpoint-session-timeout"
}

BROKER_WS_HTTP_ERR_CODE_TO_MSG = {
    429: "Reached the concurrent session limit, couldn't start Thundra debugger.",
    401: "Authentication is failed, check your Thundra debugger authentication token.",
    409: "Another session with the same session name exists, connection closed.",
}

ws = None
debugger_socket = None


def handle_error_message(error):
    if hasattr(error, 'status_code'):
        print(BROKER_WS_HTTP_ERR_CODE_TO_MSG.get(error.status_code, "Broker connection got error: {}".format(error)),
              file=sys.stderr)
    else:
        print("Broker connection got error: {}".format(error), file=sys.stderr)


def handle_close_message(code, message):
    if code is None and message is None:
        print("Thundra debug broker connection is closed")
    elif code:
        print("Thundra debug broker closed with code:{}".format(code))
    else:
        print("Thundra debug broker closed with code:{}, message:{}".format(code, message))


def on_open(ws):
    def run():
        try:
            ws.running = True
            while ws.running:
                rlist = select.select([debugger_socket, sys.stdin], [], [])[0]

                if sys.stdin in rlist:
                    line = sys.stdin.readline()
                    if line.strip() == "fin":
                        ws.running = False

                if debugger_socket in rlist:
                    buf = debugger_socket.recv(4096)
                    if len(buf) == 0:
                        ws.running = False
                    ws.send(buf, opcode=OPCODE_BINARY)

        except Exception as e:
            print("Exception while listening from debugger socket and stdin: {}".format(e))
        ws.close()

    thread.start_new_thread(run, ())


def on_message(ws, message):
    try:
        if isinstance(message, bytes):
            ws.debugger_socket.send(message)
        else:
            ws.debugger_socket.send(message.encode())
    except Exception as e:
        print("Error on on_message: {}".format(e))


def on_error(ws, error):
    handle_error_message(error)
    ws.running = False


def on_close(ws, code, message):
    handle_close_message(code, message)
    ws.running = False


def normalize_broker_host(host):
    if host.startswith("wss://") or host.startswith("ws://"):
        return host
    return "wss://" + host


try:
    debugger_socket = socket.socket()
    debugger_socket.connect(("localhost", int(os.environ.get('DEBUGGER_PORT'))))
    ws = websocket.WebSocketApp(
        "{}:{}".format(normalize_broker_host(os.environ.get('BROKER_HOST')), os.environ.get('BROKER_PORT')),
        header=[
            "{auth_token_header}: {value}".format(auth_token_header=BROKER_HANDSHAKE_HEADERS.get("AUTH_TOKEN"),
                                                  value=os.environ.get("AUTH_TOKEN")),
            "{session_name_header}: {value}".format(session_name_header=BROKER_HANDSHAKE_HEADERS.get("SESSION_NAME"),
                                                    value=os.environ.get("SESSION_NAME")),
            "{protocol_ver_header}: {value}".format(protocol_ver_header=BROKER_HANDSHAKE_HEADERS.get("PROTOCOL_VER"),
                                                    value=PROTOCOL_VER),
            "{runtime_header}: {value}".format(runtime_header=BROKER_HANDSHAKE_HEADERS.get("RUNTIME"), value="python"),
            '{session_timeout_header}: {value}'.format(
                session_timeout_header=BROKER_HANDSHAKE_HEADERS.get("SESSION_TIMEOUT"),
                value=os.environ.get("SESSION_TIMEOUT"))
        ],
        on_message=on_message,
        on_close=on_close,
        on_error=on_error
        )
    ws.debugger_socket = debugger_socket
    ws.on_open = on_open
    ws.run_forever(sockopt=((socket.IPPROTO_TCP, socket.TCP_NODELAY, 1),))

except Exception as e:
    print(e)

try:
    ws.close()
except:
    pass

try:
    debugger_socket.close()
except:
    pass



import wrapt

from thundra import constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.redis import RedisIntegration

redis_integration = RedisIntegration()


def _wrapper(wrapped, instance, args, kwargs):
    return redis_integration.run_and_trace(
        wrapped,
        instance,
        args,
        kwargs
    )


def patch():
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_REDIS_DISABLE):
        for method in map(str.lower, constants.RedisCommandTypes.keys()):
            try:
                wrapt.wrap_function_wrapper(
                    'redis.client',
                    'Redis.' + method,
                    _wrapper
                )
            except:
                pass



from __future__ import absolute_import, unicode_literals

import sys
from importlib.machinery import PathFinder, ModuleSpec, SourceFileLoader

import thundra.utils as utils
from thundra import constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.plugins.trace.traceable import Traceable


class ImportPatcher(utils.Singleton):

    def __init__(self):
        self.modules_map = self.__process_env_var_modules_to_instrument()

        for module_path in self.modules_map.keys():
            sys.meta_path.insert(0, ThundraFinder(module_path))

    @staticmethod
    def __process_env_var_modules_to_instrument():
        modules = {}
        for variable in ConfigProvider.configs:
            value = ConfigProvider.get(variable)
            if variable.startswith(config_names.THUNDRA_TRACE_INSTRUMENT_TRACEABLECONFIG):
                try:
                    module_path, function_prefix, arguments = utils.process_trace_def_var(value)
                    if module_path is not None:
                        modules[module_path] = [function_prefix, arguments]
                except:
                    pass

        return modules

    def get_module_function_prefix(self, module_name):
        try:
            return self.modules_map[module_name][0]
        except:
            return None

    def get_trace_arguments(self, module_name):
        try:
            return self.modules_map[module_name][1]
        except:
            return None


class ThundraFinder(PathFinder):

    def __init__(self, module_name):
        self.module_name = module_name

    def find_spec(self, fullname, path=None, target=None):
        if fullname == self.module_name:
            spec = super(ThundraFinder, self).find_spec(fullname, path, target)
            loader = ThundraLoader(fullname, spec.origin)
            return ModuleSpec(fullname, loader)


# Loading the module in a load time
class ThundraLoader(SourceFileLoader):

    def exec_module(self, module):
        super(ThundraLoader, self).exec_module(module)
        import_patcher = ImportPatcher()
        module_name = utils.get_module_name(module)
        function_prefix = import_patcher.get_module_function_prefix(module_name)
        trace_args_list = import_patcher.get_trace_arguments(module_name)
        try:
            trace_args = utils.str2bool(trace_args_list[constants.TRACE_ARGS])
        except:
            trace_args = False

        try:
            trace_return_value = utils.str2bool(trace_args_list[constants.TRACE_RETURN_VALUE])
        except:
            trace_return_value = False

        try:
            trace_error = utils.str2bool(trace_args_list[constants.TRACE_ERROR])
        except:
            trace_error = True

        allowed_functions = utils.get_allowed_functions(module)

        if function_prefix != '':
            for function in allowed_functions:
                if function_prefix in function:
                    setattr(module, function,
                            Traceable(trace_args=trace_args,
                                      trace_return_value=trace_return_value,
                                      trace_error=trace_error)(getattr(module, function)))
        else:
            for function in allowed_functions:
                setattr(module, function,
                        Traceable(trace_args=trace_args,
                                  trace_return_value=trace_return_value,
                                  trace_error=trace_error)(getattr(module, function)))

        return module



from thundra import constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.base_integration import BaseIntegration


class ElasticsearchIntegration(BaseIntegration):
    CLASS_TYPE = 'elasticsearch'

    def __init__(self):
        pass

    def get_hosts(self, instance):
        try:
            hosts = [con.host for con in instance.connection_pool.connections]
            return hosts
        except Exception:
            return []

    def get_normalized_path(self, es_uri):
        path_depth = ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_ELASTICSEARCH_PATH_DEPTH)

        path_seperator_count = 0
        normalized_path = ''
        prev_c = ''
        for c in es_uri:
            if c == '/' and prev_c != '/':
                path_seperator_count += 1

            if path_seperator_count > path_depth:
                break

            normalized_path += c
            prev_c = c
        return normalized_path

    def get_operation_name(self, wrapped, instance, args, kwargs):
        try:
            es_uri = args[1]
            return self.get_normalized_path(es_uri)
        except KeyError:
            return ''

    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        scope.span.class_name = constants.ClassNames['ELASTICSEARCH']
        scope.span.domain_name = constants.DomainNames['DB']

        operation_name = self.get_operation_name(wrapped, instance, args, kwargs)
        hosts = self.get_hosts(instance)

        http_method, es_path = args
        es_body = kwargs.get('body', {})
        es_params = kwargs.get('params', {})

        tags = {
            constants.ESTags['ES_HOSTS']: hosts,
            constants.ESTags['ES_URI']: es_path,
            constants.ESTags['ES_NORMALIZED_URI']: operation_name,
            constants.ESTags['ES_METHOD']: http_method,
            constants.ESTags['ES_PARAMS']: es_params,
            constants.DBTags['DB_TYPE']: 'elasticsearch',
            constants.SpanTags['OPERATION_TYPE']: http_method,
            constants.SpanTags['TOPOLOGY_VERTEX']: True,
        }

        if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_ELASTICSEARCH_BODY_MASK):
            tags[constants.ESTags['ES_BODY']] = es_body

        scope.span.tags = tags



import mock
import json

from thundra import constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.reporter import Reporter
from thundra.encoder import to_json


@mock.patch('thundra.reporter.requests')
def test_send_report_to_url(mock_requests, mock_report):
    ConfigProvider.set(config_names.THUNDRA_REPORT_REST_BASEURL, 'different_url/api')
    ConfigProvider.set(config_names.THUNDRA_REPORT_REST_COMPOSITE_ENABLE, 'false')
    test_session = mock_requests.Session()
    reporter = Reporter('api key', session=test_session)
    responses = reporter.send_reports([mock_report])

    post_url = 'different_url/api/monitoring-data'
    headers = {
        'Content-Type': 'application/json',
        'Authorization': 'ApiKey api key'
    }

    reporter.session.post.assert_called_once_with(post_url, data=to_json([mock_report], separators=(',', ':')),
                                                  headers=headers, timeout=constants.DEFAULT_REPORT_TIMEOUT)
    reporter.session.post.return_value.status_code = 200
    for response in responses:
        assert response.status_code == 200


@mock.patch('thundra.reporter.requests')
def test_send_report(mock_requests, mock_invocation_report):
    test_session = mock_requests.Session()
    reporter = Reporter('unauthorized api key', session=test_session)
    responses = reporter.send_reports([mock_invocation_report])

    assert reporter.session.post.call_count == 1
    test_session.post.return_value.status_code = 401
    for response in responses:
        assert response.status_code == 401


def test_get_report_batches(mock_report):
    ConfigProvider.set(config_names.THUNDRA_REPORT_REST_COMPOSITE_BATCH_SIZE, '2')

    reporter = Reporter('api key')
    batches = reporter.get_report_batches([mock_report] * 3)

    assert len(batches) == 2
    assert batches[0] == [mock_report, mock_report]
    assert batches[1] == [mock_report]


def test_prepare_report_json(mock_report, mock_report_with_byte_field):
    reporter = Reporter('api key')

    reports = reporter.prepare_report_json([mock_report, mock_report_with_byte_field])
    reports = json.loads(reports[0])

    assert len(reports) == 2
    assert reports[0].get('type') != 'bytes'
    assert reports[1].get('type') == 'bytes'


def test_prepare_report_json_batch(mock_report):
    ConfigProvider.set(config_names.THUNDRA_REPORT_REST_COMPOSITE_BATCH_SIZE, '1')

    reporter = Reporter('api key')

    batched_reports = reporter.prepare_report_json([mock_report] * 2)
    assert len(batched_reports) == 2

    reports = json.loads(batched_reports[0])
    assert len(reports) == 1



import mock
import pytest

from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.context.execution_context_manager import ExecutionContextManager
from thundra.plugins.trace.trace_plugin import TracePlugin
from thundra.thundra_agent import Thundra


def test_if_api_key_is_retrieved_from_env_var():
    ConfigProvider.set(config_names.THUNDRA_APIKEY, 'api key')
    thundra = Thundra()
    assert thundra.api_key == 'api key'


def test_if_disable_trace_is_set_to_true():
    thundra = Thundra('api key', disable_trace=True)

    for plugin in thundra.plugins:
        assert not type(plugin) is TracePlugin


def test_if_disable_trace_is_set_to_false():
    thundra = Thundra('api key', disable_trace=False)

    trace_exist = False
    for plugin in thundra.plugins:
        if isinstance(plugin, TracePlugin):
            trace_exist = True

    assert trace_exist is True


def test_if_disable_trace_is_not_set():
    thundra = Thundra('api key')

    trace_exist = False
    for plugin in thundra.plugins:
        if isinstance(plugin, TracePlugin):
            trace_exist = True

    assert trace_exist is True


def test_disable_trace_plugin_from_environment_variable():
    ConfigProvider.set(config_names.THUNDRA_TRACE_DISABLE, 'true')
    thundra = Thundra('api key')

    trace_exist = False
    for plugin in thundra.plugins:
        if isinstance(plugin, TracePlugin):
            trace_exist = True

    assert trace_exist is False


def test_enable_trace_plugin_from_environment_variable():
    ConfigProvider.set(config_names.THUNDRA_TRACE_DISABLE, 'false')
    thundra = Thundra('api key')

    trace_exist = False
    for plugin in thundra.plugins:
        if isinstance(plugin, TracePlugin):
            trace_exist = True

    assert trace_exist is True


def test_if_disable_trace_plugin_from_environment_variable_is_prior():
    ConfigProvider.set(config_names.THUNDRA_TRACE_DISABLE, 'true')
    thundra = Thundra('api key', disable_trace=False)

    trace_exist = False
    for plugin in thundra.plugins:
        if isinstance(plugin, TracePlugin):
            trace_exist = True

    assert trace_exist is False


def test_if_enable_trace_plugin_from_environment_variable_is_prior():
    ConfigProvider.set(config_names.THUNDRA_TRACE_DISABLE, 'false')
    thundra = Thundra('api key', disable_trace=True)

    trace_exist = False
    for plugin in thundra.plugins:
        if isinstance(plugin, TracePlugin):
            trace_exist = True

    assert trace_exist is True


@mock.patch('thundra.reporter.Reporter')
def test_if_thundra_is_disabled(mock_reporter, handler, mock_event, mock_context):
    ConfigProvider.set(config_names.THUNDRA_TRACE_DISABLE, 'true')
    _, handler = handler

    handler(mock_event, mock_context)

    assert not mock_reporter.add_report.called
    assert not mock_reporter.send_reports.called


def test_if_exception_is_handled(handler_with_exception, mock_context, mock_event):
    thundra, handler = handler_with_exception
    with pytest.raises(Exception):
        handler(mock_event, mock_context)

    assert ExecutionContextManager.get().error


@mock.patch('thundra.thundra_agent.Thundra.check_and_handle_warmup_request')
def test_if_thundra_crashes_user_handler_before(mocked_func, handler, mock_event, mock_context):
    mocked_func.side_effect = RuntimeError('Boom!')
    thundra, handler = handler
    try:
        handler(mock_event, mock_context)
    except Exception:
        pytest.fail("User's handler shouldn't fail when Thundra raise an exception")


@mock.patch('thundra.reporter.Reporter.send_reports')
def test_if_thundra_crashes_user_handler_after(mocked_func, handler, mock_event, mock_context):
    mocked_func.side_effect = RuntimeError('Boom!')
    thundra, handler = handler
    try:
        handler(mock_event, mock_context)
    except Exception:
        pytest.fail("User's handler shouldn't fail when Thundra raise an exception")
